from __future__ import annotations

# packages
import numpy as np
import pandas as pd
import seaborn as sns
import matplotlib.pyplot as plt
from numba import njit
from dataclasses import dataclass
from typing import Union, Dict, Any, Optional, Tuple, List
from enum import Enum

# qis
import qis as qis
import qis.utils.annualisation
from qis import TimePeriod, PerfStat, PerfParams, RegimeData, EnumMap, BenchmarkReturnsQuantilesRegime, RollingPerfStat
import qis.utils.df_groups as dfg
import qis.utils.df_agg as dfa
import qis.perfstats.returns as ret
import qis.perfstats.perf_stats as rpt
import qis.perfstats.regime_classifier as rcl
import qis.plots.time_series as pts
import qis.plots.stackplot as pst
import qis.plots.derived.prices as ppd
import qis.plots.derived.perf_table as ppt
import qis.plots.derived.returns_scatter as prs
import qis.plots.derived.returns_heatmap as rhe
from qis.models.linear.ewm import compute_ewm_vol
import qis.portfolio.risk.ewm_factor_model as ef
from qis.portfolio.signal_data import StrategySignalData
from qis.portfolio.risk.ewm_covar_risk import compute_portfolio_vol
from qis.portfolio.risk.contributions import compute_portfolio_risk_contributions

# default performance and regime params
PERF_PARAMS = PerfParams(freq='W-WED')
regime_classifier = BenchmarkReturnsQuantilesRegime(freq='ME')


class AttributionMetric(str, Enum):
    """
    input for computation of get_performance_attribution_data()
    """
    PNL = 'P&L Attribution, sum=portfolio performance'
    PNL_RISK = 'P&L Risk Attribution, sum=100%'
    INST_PNL = 'Instrument P&L'
    COSTS = 'Instrument Total Costs'
    TURNOVER = 'Instrument Annualised Turnover'
    VOL_ADJUSTED_TURNOVER = 'Instrument Annualised Vol-Adjusted Turnover'


class SnapshotPeriod(str, Enum):
    """
    input for computation of get_performance_attribution_data()
    """
    LAST = 'last'
    AVG = 'avg'
    MAX = 'max'


@dataclass
class PortfolioData:
    """
    portfolio data can be generated by
    1. qis.backtest_model_portfolio()
    2. independent backtester with outputs matching PortfolioData inputs
    it is possible to supply StrategySignalData for 2
    """
    nav: pd.Series  # nav is computed with all cost
    weights: pd.DataFrame = None  # weights of portfolio
    units: pd.DataFrame = None  # units of portfolio instruments
    prices: pd.DataFrame = None  # prices of portfolio universe
    instrument_pnl: pd.DataFrame = None  # include net pnl by instrument
    realized_costs: pd.DataFrame = None  # realized trading costs by instrument
    input_weights: Union[np.ndarray, pd.DataFrame, Dict[str, float]] = None  # inputs to potfolio
    is_rebalancing: pd.Series = None  # optional info when the weights are rebalanced
    tickers_to_names_map: Optional[Dict[str, str]] = None  # renaming of long tickers
    group_data: pd.Series = None  # for asset class grouping
    group_order: List[str] = None
    instrument_names: pd.Series = None
    benchmark_prices: pd.DataFrame = None  # can pass benchmark prices here
    ticker: str = None
    strategy_signal_data: StrategySignalData = None
    covar_dict: Dict[pd.Timestamp, pd.DataFrame] = None  # for computing risk contributions

    def __post_init__(self):
        if isinstance(self.nav, pd.DataFrame):
            self.nav = self.nav.iloc[:, 0]
        if self.prices is None:
            self.prices = self.nav.to_frame()
        if self.weights is None:  # default will be delta-1 portfolio of nav
            self.weights = pd.DataFrame(1.0, index=self.prices.index, columns=self.prices.columns)
        if self.units is None:  # default will be delta-1 portfolio of nav
            self.units = pd.DataFrame(1.0, index=self.prices.index, columns=self.prices.columns)
        if self.realized_costs is None:
            self.realized_costs = pd.DataFrame(0.0, index=self.prices.index, columns=self.prices.columns)
        if self.instrument_pnl is None:
            self.instrument_pnl = self.prices.pct_change(fill_method=None).multiply(self.weights.shift(1)).fillna(0.0)
        if self.group_data is None:  # use instruments as groups
            self.group_data = pd.Series(self.prices.columns, index=self.prices.columns)
        if self.group_order is None:
            self.group_order = list(self.group_data.unique())
        if self.benchmark_prices is not None:
            self.benchmark_prices = self.benchmark_prices.reindex(index=self.nav.index, method='ffill')
        if self.ticker is None:
            self.ticker = str(self.nav.name)

    def set_ticker(self, ticker: str) -> PortfolioData:
        self.ticker = ticker
        self.nav = self.nav.rename(ticker)
        return self

    def set_benchmark_prices(self, benchmark_prices: Union[pd.Series, pd.DataFrame]) -> None:
        # can pass benchmark prices here
        if isinstance(benchmark_prices, pd.Series):
            benchmark_prices = benchmark_prices.to_frame()
        self.benchmark_prices = benchmark_prices.reindex(index=self.nav.index, method='ffill')

    def set_group_data(self, group_data: pd.Series, group_order: List[str] = None) -> None:
        self.group_data = group_data
        if group_order is None:
            group_order = group_data.to_list()
        self.group_order = group_order

    def save(self, ticker: str, local_path: str = './') -> None:
        datasets = dict(nav=self.nav, prices=self.prices, weights=self.weights, units=self.units,
                        instrument_pnl=self.instrument_pnl, realized_costs=self.realized_costs)
        if self.group_data is not None:
            datasets['group_data'] = self.group_data
        qis.save_df_dict_to_csv(datasets=datasets, file_name=ticker, local_path=local_path)
        print(f"saved portfolio data for {ticker}")

    @classmethod
    def load(cls, ticker: str) -> PortfolioData:
        dataset_keys = ['nav', 'prices', 'weights', 'units', 'instrument_pnl', 'realized_costs', 'group_data']
        datasets = qis.load_df_dict_from_csv(dataset_keys=dataset_keys, file_name=ticker)
        return cls(**datasets)

    """
    NAV level getters
    """

    def get_portfolio_nav(self, time_period: TimePeriod = None, freq: Optional[str] = None,
                          ticker: Optional[str] = None) -> pd.Series:
        """
        get nav using consistent function for all return computations
        """
        if time_period is not None:
            nav_ = time_period.locate(self.nav)
        else:
            nav_ = self.nav.copy()
        if freq is not None:
            nav_ = nav_.asfreq(freq=freq, method='ffill')
        if ticker is not None:
            nav_ = nav_.rename(ticker)
        return nav_

    def get_portfolio_nav_with_benchmark_prices(self,
                                                time_period: TimePeriod = None,
                                                freq: Optional[str] = None
                                                ) -> pd.DataFrame:
        """
        get nav using consistent function for all return computations
        """
        navs = self.get_portfolio_nav(time_period=time_period, freq=freq)
        if self.benchmark_prices is not None:
            benchmark_prices = self.benchmark_prices.reindex(index=navs.index, method='ffill')
            navs = pd.concat([navs, benchmark_prices], axis=1)
        return navs

    def get_instruments_pnl(self,
                            add_total: bool = False,
                            time_period: TimePeriod = None,
                            is_net: bool = False,
                            is_unit_based_traded_volume: bool = True,
                            is_compounded: bool = False,
                            freq: Optional[str] = None
                            ) -> pd.DataFrame:
        pnl = self.instrument_pnl.copy()
        if is_net:
            costs = self.get_costs(add_total=False, is_unit_based_traded_volume=is_unit_based_traded_volume)
            pnl = pnl.subtract(costs)
        if add_total:
            pnl.insert(loc=0, value=pnl.sum(1), column='Total')
        if time_period is not None:
            pnl = time_period.locate(pnl)
        if freq is not None:
            pnl = pnl.resample(freq).sum()
        if is_compounded:
            pnl = np.expm1(pnl)
        return pnl

    def get_performance_attribution(self,
                                    add_total: bool = True,
                                    time_period: TimePeriod = None,
                                    is_compounded: bool = False
                                    ) -> pd.Series:
        instrument_pnl = self.get_instruments_pnl(add_total=add_total,
                                                  time_period=time_period,
                                                  is_compounded=is_compounded)
        if is_compounded:
            performance_attribution = instrument_pnl.iloc[-1, :] - 1.0
        else:
            performance_attribution = instrument_pnl.sum(axis=0)
        return performance_attribution

    def get_instruments_navs(self,
                             time_period: TimePeriod = None,
                             constant_trade_level: bool = False
                             ) -> pd.DataFrame:
        pnl = self.get_instruments_pnl(time_period=time_period, is_compounded=False).fillna(0.0)
        navs = ret.returns_to_nav(returns=pnl, constant_trade_level=constant_trade_level)
        return navs

    def get_group_navs(self,
                       time_period: TimePeriod = None,
                       constant_trade_level: bool = False,
                       is_add_group_total: bool = False
                       ) -> Optional[pd.DataFrame]:
        """
        group total will exclude transaction costs so it is not equal to portfolio nav
        """
        if is_add_group_total:
            total_column = str(self.nav.name)
        else:
            total_column = None
        df = self.get_instruments_pnl(time_period=time_period)
        if df.empty:
            print(f"instruments p&l is not available for time_period={time_period.to_str()}")
            group_navs = None
        else:
            grouped_pnl = dfg.agg_df_by_groups_ax1(df=df,
                                                   group_data=self.group_data,
                                                   agg_func=np.nansum,
                                                   total_column=total_column,
                                                   group_order=self.group_order)
            group_navs = ret.returns_to_nav(returns=grouped_pnl, constant_trade_level=constant_trade_level)
        return group_navs

    def get_total_nav_with_group_navs(self, time_period: TimePeriod = None) -> pd.DataFrame:
        """
        group total will exclude transaction costs so it is not equal to portfolio nav
        """
        total_nav = self.get_portfolio_nav(time_period=time_period)
        group_navs = self.get_group_navs(time_period=time_period, is_add_group_total=False)
        prices = pd.concat([total_nav, group_navs], axis=1)
        return prices

    def get_input_weights(self, time_period: TimePeriod = None) -> Union[np.ndarray, pd.DataFrame, Dict[str, float]]:
        input_weights = self.input_weights.copy()
        if time_period is not None and isinstance(input_weights, pd.DataFrame):
            input_weights = time_period.locate(input_weights)
        return input_weights

    def get_weights(self,
                    is_input_weights: bool = False,
                    columns: List[str] = None,
                    time_period: TimePeriod = None,
                    is_grouped: bool = False,
                    add_total: bool = False,
                    freq: Optional[str] = 'W-WED',
                    group_data: pd.Series = None,
                    group_order: List[str] = None
                    ) -> pd.DataFrame:
        if is_input_weights and self.input_weights is not None and isinstance(self.input_weights, pd.DataFrame):
            weights = self.input_weights.copy()
        else:
            weights = self.weights.copy()
        if freq is not None:
            weights = weights.resample(freq).last().ffill()

        if is_grouped:
            if group_data is None:
                group_data = self.group_data
            if group_order is None:
                group_order = self.group_order
            weights = dfg.agg_df_by_groups_ax1(df=weights,
                                               group_data=group_data,
                                               agg_func=np.nansum,
                                               total_column=self.ticker if add_total else None,
                                               group_order=group_order)
        elif columns is not None:
            weights = weights[columns]

        if time_period is not None:
            weights = time_period.locate(weights)

        return weights

    def get_grouped_long_short_weights(self,
                                       time_period: TimePeriod = None,
                                       total_name: str = 'Total',
                                       weights_freq: Optional[str] = 'W-WED',
                                       total_column: Optional[str] = None
                                       ) -> Tuple[Dict[str, pd.DataFrame], Dict[str, pd.DataFrame]]:
        """Computes grouped net long/short exposures by instrument groups.

        Groups portfolio weights by predefined instrument categories and calculates
        aggregate exposure metrics including total exposure, net long positions, and
        net short positions for each group.

        Args:
            time_period: Time period for which to calculate exposures. If None,
                uses the full available time series.
            total_name: Column name for the total exposure aggregation. 
                Defaults to 'Total'.
            weights_freq: Frequency for resampling exposures data. 
                Defaults to 'W-WED' (weekly on Wednesday).
            total_column: Column name for total in group dictionary. If None,
                no total column is included in grouping.
        Returns:
            A tuple containing:
                - grouped_weights_agg: Dictionary mapping group names to DataFrames
                  with columns for total, net long, and net short exposures.
                - grouped_weights_by_inst: Dictionary mapping group names to 
                  DataFrames containing individual instrument exposures within 
                  each group.

        Note:
            Positive exposures represent long positions, negative exposures 
            represent short positions. Net long sums only positive values,
            net short sums only negative values.
        """
        group_dict = dfg.get_group_dict(group_data=self.group_data,
                                        group_order=self.group_order,
                                        total_column=total_column)
        all_exposures = self.get_weights(time_period=time_period, freq=weights_freq)
        grouped_weights_by_inst = {}
        grouped_weights_agg = {}
        for group, tickers in group_dict.items():
            exposures_by_inst = all_exposures[tickers]
            grouped_weights_by_inst[group] = exposures_by_inst
            total = dfa.nansum(exposures_by_inst, axis=1).rename(total_name)
            net_long = dfa.nansum_positive(exposures_by_inst, axis=1).rename('Net Long')
            net_short = dfa.nansum_negative(exposures_by_inst, axis=1).rename('Net Short')
            grouped_weights_agg[group] = pd.concat([total, net_long, net_short], axis=1)
        return grouped_weights_agg, grouped_weights_by_inst

    def get_grouped_cum_pnls(self,
                             total_name: str = 'Total',
                             time_period: TimePeriod = None
                             ) -> Tuple[Dict[str, pd.DataFrame], Dict[str, pd.DataFrame]]:
        """
        compute grouped net long / short exposues
        """
        group_dict = dfg.get_group_dict(group_data=self.group_data,
                                        group_order=self.group_order,
                                        total_column=None)

        pnls = self.get_instruments_pnl(time_period=time_period, is_compounded=False).fillna(0.0)
        all_exposures = self.get_weights(time_period=time_period, freq=None)
        all_exposures = all_exposures.reindex(index=pnls.index, method='ffill').ffill()
        pnl_positive_exp = pnls.where(all_exposures > 0.0, other=0.0)
        pnl_negative_exp = pnls.where(all_exposures < 0.0, other=0.0)

        grouped_pnls_by_inst = {}
        grouped_pnls_agg = {}
        for group, tickers in group_dict.items():
            pnls_by_inst = pnls[tickers]
            grouped_pnls_by_inst[group] = pnls_by_inst.cumsum(axis=0)
            total = dfa.nansum(pnls_by_inst, axis=1).rename(total_name)
            net_long = dfa.nansum(pnl_positive_exp[tickers], axis=1).rename('Net Long')
            net_short = dfa.nansum(pnl_negative_exp[tickers], axis=1).rename('Net Short')
            grouped_pnls_agg[group] = pd.concat([total, net_long, net_short], axis=1).cumsum(axis=0)
        return grouped_pnls_agg, grouped_pnls_by_inst

    def get_turnover(self,
                     is_agg: bool = False,
                     is_grouped: bool = False,
                     group_data: pd.Series = None,
                     group_order: List[str] = None,
                     time_period: TimePeriod = None,
                     roll_period: Optional[int] = 260,
                     is_vol_adjusted: bool = False,
                     add_total: bool = True,
                     vol_span: int = 33,
                     freq: Optional[str] = None,
                     is_unit_based_traded_volume: bool = True,
                     **kwargs
                     ) -> Union[pd.DataFrame, pd.Series]:

        if is_unit_based_traded_volume:  # for unit generated backtest
            turnover = (self.units.diff(1).abs()).multiply(self.prices)
            abs_exposure = self.units.multiply(self.prices).abs().sum(axis=1)
            # turnover = turnover.divide(self.nav.to_numpy(), axis=0)
            turnover = turnover.divide(abs_exposure.to_numpy(), axis=0)
        else:  # for weight generated backtets
            turnover = self.input_weights.diff(1).abs()

        if is_vol_adjusted:
            instrument_vols = compute_ewm_vol(data=qis.to_returns(self.prices, is_log_returns=True),
                                              span=vol_span,
                                              annualize=True)
            turnover = turnover.multiply(instrument_vols)

        if is_agg:
            turnover = pd.Series(np.nansum(turnover, axis=1), index=turnover.index, name=self.nav.name)
            turnover = turnover.reindex(index=self.nav.index)
        elif is_grouped:  # agg by groups
            if group_data is None:
                group_data = self.group_data
            if group_order is None:
                group_order = self.group_order
            turnover = dfg.agg_df_by_groups_ax1(df=turnover,
                                                group_data=group_data,
                                                agg_func=np.nansum,
                                                total_column=str(self.nav.name) if add_total else None,
                                                group_order=group_order)
        else:
            if add_total:
                turnover = pd.concat([turnover.sum(axis=1).rename(self.nav.name), turnover], axis=1)

        if not turnover.empty:  # it may happen for undefined groupings
            if freq is not None:  # first aggregate by freq
                turnover = turnover.resample(freq).sum()
            if roll_period is not None:  # now aggregate by roll
                turnover = turnover.rolling(roll_period).sum()
            if time_period is not None:
                turnover = time_period.locate(turnover)
        return turnover

    def get_costs(self,
                  is_agg: bool = False,
                  is_grouped: bool = False,
                  time_period: TimePeriod = None,
                  add_total: bool = True,
                  is_unit_based_traded_volume: bool = True,
                  roll_period: Optional[int] = 260,
                  freq: Optional[str] = None
                  ) -> Union[pd.DataFrame, pd.Series]:

        costs = self.realized_costs
        if is_unit_based_traded_volume:
            costs = costs.divide(self.nav.to_numpy(), axis=0)
        if is_agg:
            costs = pd.Series(np.nansum(costs, axis=1), index=self.nav.index, name=self.nav.name)
        elif is_grouped:  # agg by groups
            costs = dfg.agg_df_by_groups_ax1(costs,
                                             group_data=self.group_data,
                                             agg_func=np.nansum,
                                             total_column=str(self.nav.name) if add_total else None,
                                             group_order=self.group_order)
        else:
            if add_total:
                costs = pd.concat([costs.sum(axis=1).rename(self.nav.name), costs], axis=1)

        if freq is not None:  # first aggregate by freq
            costs = costs.resample(freq).sum()
        if roll_period is not None:  # now aggregate by roll period
            costs = costs.rolling(roll_period).sum()
        if time_period is not None:
            costs = time_period.locate(costs)
        return costs

    def compute_mcap_participation(self,
                                   mcap: pd.DataFrame,
                                   trade_level: float = 100000000
                                   ) -> pd.DataFrame:
        exposure = (self.units.multiply(self.prices)).divide(self.nav.to_numpy(), axis=0)
        participation = trade_level * exposure.divide(mcap)
        return participation

    def compute_volume_participation(self,
                                     volumes: pd.DataFrame,
                                     trade_level: float = 100000000
                                     ) -> pd.DataFrame:
        turnover = self.get_turnover(is_agg=False, is_grouped=False)
        participation = trade_level * turnover.divide(volumes)
        return participation

    def compute_cumulative_attribution(self) -> pd.DataFrame:
        attribution = (self.prices.pct_change()).multiply(self.weights.shift(1))
        attribution = attribution.cumsum(axis=0)
        return attribution

    def compute_realized_pnl(self, time_period: TimePeriod = None) -> Tuple[pd.DataFrame, ...]:
        avg_costs, realized_pnl, mtm_pnl, trades = compute_realized_pnl(prices=self.prices.to_numpy(),
                                                                        units=self.units.to_numpy())
        avg_costs = pd.DataFrame(avg_costs, index=self.prices.index, columns=self.prices.columns)
        realized_pnl = pd.DataFrame(realized_pnl, index=self.prices.index, columns=self.prices.columns)
        mtm_pnl = pd.DataFrame(mtm_pnl, index=self.prices.index, columns=self.prices.columns)
        trades = pd.DataFrame(trades, index=self.prices.index, columns=self.prices.columns)
        if time_period is not None:
            avg_costs = time_period.locate(avg_costs)
            realized_pnl = time_period.locate(realized_pnl)
            mtm_pnl = time_period.locate(mtm_pnl)
            trades = time_period.locate(trades)
        realized_pnl = realized_pnl.cumsum(axis=0)
        total_pnl = realized_pnl.add(mtm_pnl)
        return avg_costs, realized_pnl, mtm_pnl, total_pnl, trades

    def compute_portfolio_benchmark_betas(self,
                                          benchmark_prices: pd.DataFrame,
                                          time_period: TimePeriod = None,
                                          freq_beta: str = None,
                                          factor_beta_span: int = 65  # quarter
                                          ) -> pd.DataFrame:
        """
        compute benchmark betas of instruments
        portfolio_beta_i = sum(instrument_beta_i*exposure)
        """
        instrument_prices = self.prices
        benchmark_prices = benchmark_prices.reindex(index=instrument_prices.index, method='ffill')
        exposures = self.get_weights().reindex(index=instrument_prices.index, method='ffill')
        benchmark_betas = ef.compute_portfolio_ewm_benchmark_betas(instrument_prices=instrument_prices,
                                                                   weights=exposures,
                                                                   benchmark_prices=benchmark_prices,
                                                                   time_period=time_period,
                                                                   freq_beta=freq_beta,
                                                                   factor_beta_span=factor_beta_span)
        return benchmark_betas

    def compute_portfolio_benchmark_attribution(self,
                                                benchmark_prices: pd.DataFrame,
                                                time_period: TimePeriod = None,
                                                freq_beta: str = 'B',
                                                factor_beta_span: int = 63,  # quarter
                                                residual_name: str = 'Alpha'
                                                ) -> pd.DataFrame:
        """
        attribution = portfolio_return_{t} - benchmark_return_{t}*bet_{t-1}
        returns are compounded
        """
        instrument_prices = self.prices
        benchmark_prices = benchmark_prices.reindex(index=instrument_prices.index, method='ffill')
        weights = self.get_weights().reindex(index=instrument_prices.index, method='ffill')
        portfolio_nav = self.get_portfolio_nav().reindex(index=instrument_prices.index, method='ffill')
        joint_attrib = ef.compute_portfolio_benchmark_ewm_beta_alpha_attribution(instrument_prices=instrument_prices,
                                                                                 weights=weights,
                                                                                 benchmark_prices=benchmark_prices,
                                                                                 portfolio_nav=portfolio_nav,
                                                                                 time_period=time_period,
                                                                                 freq_beta=freq_beta,
                                                                                 factor_beta_span=factor_beta_span,
                                                                                 residual_name=residual_name)
        return joint_attrib

    """
    ### instrument level getters
    """

    def get_instruments_returns(self,
                                time_period: TimePeriod = None
                                ) -> pd.DataFrame:
        returns = self.prices.pct_change(fill_method=None)
        if time_period is not None:
            returns = time_period.locate(returns)
        return returns

    def get_instruments_periodic_returns(self,
                                         time_period: TimePeriod = None,
                                         freq: str = 'ME'
                                         ) -> pd.DataFrame:
        returns = self.get_instruments_returns(time_period=time_period)
        prices = ret.returns_to_nav(returns=returns, init_period=None)
        returns_f = ret.to_returns(prices=prices, freq=freq)
        return returns_f

    def get_instruments_performance_attribution(self,
                                                time_period: TimePeriod = None,
                                                constant_trade_level: bool = False
                                                ) -> pd.Series:
        navs = self.get_instruments_navs(time_period=time_period, constant_trade_level=constant_trade_level)
        perf = ret.to_total_returns(prices=navs).rename(self.nav.name)
        return perf

    def get_instruments_pnl_risk_attribution(self,
                                             time_period: TimePeriod = None
                                             ) -> pd.Series:
        pnl = self.get_instruments_pnl(time_period=time_period)
        # portfolio_pnl = pnl.sum(axis=1)

        pnl_risk = np.nanstd(pnl.replace({0.0: np.nan}), axis=0)
        # portfolio_pnl_risk = np.nanstd(portfolio_pnl.replace({0.0: np.nan}), axis=0)
        # pnl_risk_ratio = pnl_risk / portfolio_pnl_risk
        pnl_risk_ratio = pnl_risk / np.nansum(pnl_risk)

        data = pd.Series(pnl_risk_ratio, index=pnl.columns, name=self.nav.name)
        if self.tickers_to_names_map is not None:
            data = data.rename(index=self.tickers_to_names_map)
        return data

    def get_performance_attribution_data(self,
                                         attribution_metric: AttributionMetric = AttributionMetric.PNL,
                                         time_period: TimePeriod = None,
                                         is_unit_based_traded_volume: bool = True,
                                         **kwargs
                                         ) -> Union[pd.DataFrame, pd.Series]:
        if attribution_metric == AttributionMetric.PNL:
            data = self.get_instruments_performance_attribution(time_period=time_period)
        elif attribution_metric == AttributionMetric.PNL_RISK:
            data = self.get_instruments_pnl_risk_attribution(time_period=time_period)
        elif attribution_metric == AttributionMetric.INST_PNL:
            data = self.get_instruments_navs(time_period=time_period)
        elif attribution_metric == AttributionMetric.COSTS:
            data = self.get_costs(is_agg=False, is_grouped=False, roll_period=None,
                                  is_unit_based_traded_volume=is_unit_based_traded_volume,
                                  add_total=False,
                                  time_period=time_period)
            data = data.sum(0)
            # print(f"total costs, is_unit_based_traded_volume={is_unit_based_traded_volume} = {np.nansum(data)}")
        elif attribution_metric == AttributionMetric.TURNOVER:
            data = self.get_turnover(is_agg=False, is_grouped=False, roll_period=None,
                                     add_total=False,
                                     time_period=time_period)
            an = qis.utils.annualisation.infer_an_from_data(data=data)
            data = an * data.mean(0)
            # print(f"total turnover = {np.nansum(data)}")
        elif attribution_metric == AttributionMetric.VOL_ADJUSTED_TURNOVER:
            data = self.get_turnover(is_agg=False, is_grouped=False, roll_period=None,
                                     add_total=False, is_vol_adjusted=True,
                                     time_period=time_period)
            an = qis.utils.annualisation.infer_an_from_data(data=data)
            data = an * data.mean(0)

        else:
            raise NotImplementedError(f"{attribution_metric}")

        return data

    def get_num_investable_instruments(self, time_period: TimePeriod = None) -> pd.DataFrame:
        num_available_prices = np.sum(np.where(np.isfinite(self.prices), 1.0, 0.0), axis=1)
        num_available_prices = pd.Series(num_available_prices, index=self.weights.index, name='Investable')
        non_zero_exposures = self.weights.where(np.isclose(self.weights, 0.0) == False, other=np.nan)
        num_invested_instruments = np.sum(np.where(np.isfinite(non_zero_exposures), 1.0, 0.0), axis=1)
        num_invested_instruments = pd.Series(num_invested_instruments, index=self.weights.index, name='Invested')
        df = pd.concat([num_available_prices, num_invested_instruments], axis=1)
        if time_period is not None:
            df = time_period.locate(df)
        return df

    def get_instruments_performance_table(self,
                                          time_period: TimePeriod = None,
                                          portfolio_name: str = 'Attribution'
                                          ) -> pd.DataFrame:
        """
        using avg weight
        """
        insts_returns = self.get_instruments_returns(time_period=time_period)
        insts_return = ret.to_total_returns(prices=ret.returns_to_nav(returns=insts_returns))
        weight = self.weights
        if time_period is not None:
            weight = time_period.locate(weight)
        weight = weight.mean(axis=0)
        portf_return = insts_return.multiply(weight).replace({0.0: np.nan}).dropna()
        data = pd.concat([weight.rename('Weight'),
                          insts_return.rename('Asset'),
                          portf_return.rename(portfolio_name)],
                         axis=1).dropna()
        data = data.sort_values('Weight', ascending=False)
        if self.tickers_to_names_map is not None:
            data = data.rename(index=self.tickers_to_names_map)

        return data

    def get_attribution_table_by_instrument(self,
                                            time_period: TimePeriod = None,
                                            freq: str = 'ME',
                                            is_input_weights: bool = False
                                            ) -> pd.DataFrame:
        """
        using avg weight
        """
        if is_input_weights:
            weights = self.input_weights
            if not isinstance(weights, pd.DataFrame):
                raise ValueError(f"input weights must be pd.Dataframe for is_input_weights=True")
            prices_w = self.prices.reindex(index=weights.index, method='ffill')
            returns_f = prices_w.pct_change()
            portf_return = returns_f.multiply(weights.shift(1)).iloc[1:, :]
        else:
            returns_f = self.get_instruments_periodic_returns(time_period=time_period, freq=freq)
            weight = self.weights.reindex(index=returns_f.index, method='ffill')
            # first row is None
            portf_return = returns_f.multiply(weight.shift(1)).iloc[1:, :]
        if self.tickers_to_names_map is not None:
            portf_return = portf_return.rename(columns=self.tickers_to_names_map)
        return portf_return

    def compute_portfolio_vol(self,
                              time_period: TimePeriod = None,
                              freq: str = 'W-WED',
                              span: int = 13  # 3m span of weekly returns
                              ) -> pd.DataFrame:
        """
        compute_portfolio_vol using 1) instrument weights and covar matrix 2) realised returns vol
        """
        returns_f = self.get_instruments_periodic_returns(freq=freq)
        weights = self.weights.reindex(index=returns_f.index, method='ffill')
        portfolio_vol = compute_portfolio_vol(returns=returns_f,
                                              weights=weights,
                                              span=span,
                                              annualize=True)
        strategy_vol = compute_ewm_vol(data=qis.to_returns(self.get_portfolio_nav(freq=freq), is_log_returns=True),
                                       span=span,
                                       annualize=True)
        df = pd.concat([portfolio_vol.rename('instrument weighted vol'),
                        strategy_vol.rename('strategy returns vol')
                        ], axis=1)
        if time_period is not None:
            df = time_period.locate(df)
        return df

    def compute_distribution_yield(self,
                                   paid_dividends: pd.DataFrame,
                                   div_rolling_freq: str = 'ME',
                                   div_rolling_period: int = 12,
                                   time_period: TimePeriod = None
                                   ) -> Tuple[pd.DataFrame, pd.Series, pd.Series]:
        positions = self.units
        distributions_by_instrument = paid_dividends.multiply(positions)
        nav = self.get_portfolio_nav().reindex(index=distributions_by_instrument.index, method='ffill')
        distributions_by_instrument_yield = distributions_by_instrument.divide(nav, axis=0)
        # resample at 'ME' and compute rolling sum
        distributions_by_instrument_yield_freq = distributions_by_instrument_yield.resample(div_rolling_freq).sum()
        distributions_by_instrument_yield_rolling = distributions_by_instrument_yield_freq.fillna(0.0).rolling(
            div_rolling_period).sum()
        # find total distributions
        distribution_yield = distributions_by_instrument_yield_freq.sum(1)
        distribution_yield_rolling = distributions_by_instrument_yield_rolling.sum(1)
        if time_period is not None:
            distributions_by_instrument_yield_rolling = time_period.locate(distributions_by_instrument_yield_rolling)
            distribution_yield = time_period.locate(distribution_yield)
            distribution_yield_rolling = time_period.locate(distribution_yield_rolling)
        return distributions_by_instrument_yield_rolling, distribution_yield, distribution_yield_rolling

    def compute_portfolio_vars(self,
                               is_correlated: bool = True,
                               time_period: TimePeriod = None,
                               freq: str = 'B',
                               total_column: Optional[str] = 'Total',
                               vol_span: Union[int, float] = 33,  # span in number of freq-returns
                               group_data: pd.Series = None,
                               group_order: List[str] = None
                               ) -> Tuple[pd.DataFrame, Optional[pd.DataFrame]]:
        if group_data is None:
            group_data = self.group_data
        if group_order is None:
            group_order = self.group_order
        if is_correlated:
            portfolio_vars = qis.compute_portfolio_correlated_var_by_groups(prices=self.prices,
                                                                            weights=self.get_weights(freq=freq),
                                                                            group_data=group_data,
                                                                            group_order=group_order,
                                                                            freq=freq,
                                                                            vol_span=vol_span,
                                                                            # span in number of freq-returns
                                                                            total_column=total_column,
                                                                            time_period=time_period)
            instrument_vars = None
        else:
            instrument_vars, portfolio_vars = qis.compute_portfolio_independent_var_by_ac(prices=self.prices,
                                                                                          weights=self.get_weights(freq=freq),
                                                                                          group_data=group_data,
                                                                                          group_order=group_order,
                                                                                          freq=freq,
                                                                                          vol_span=vol_span,
                                                                                          # span in number of freq-returns
                                                                                          total_column=total_column,
                                                                                          time_period=time_period)
        return portfolio_vars, instrument_vars

    def compute_ex_anti_portfolio_vol_implied_by_covar(self,
                                                       covar_dict: Dict[pd.Timestamp, pd.DataFrame] = None,
                                                       freq: Optional[str] = None
                                                       ) -> pd.Series:
        """
        compute portfolio ex-anti portfolio vol using covar_dict
        """
        if covar_dict is None:
            if self.covar_dict is None:
                raise ValueError(f"must provide covar_dict")
            else:
                covar_dict = self.covar_dict
        is_input_weights = True if freq is None else False
        strategy_weights = self.get_weights(freq=freq, is_input_weights=is_input_weights)
        covar_index = list(covar_dict.keys())
        portfolio_vol = {}
        if freq is None:  # align with covar
            strategy_weights = strategy_weights.reindex(index=covar_index).ffill().fillna(0.0)
            for date, pd_covar in covar_dict.items():
                # align with covar matrix
                w = strategy_weights.loc[date].reindex(index=pd_covar.columns).fillna(0.0)
                portfolio_vol[date] = np.sqrt(w.T @ pd_covar @ w)
        else:
            for date, weights in strategy_weights.to_dict(orient='index').items():
                last_covar_update_date = qis.find_upto_date_from_datetime_index(index=covar_index, date=date)
                if last_covar_update_date is not None:
                    pd_covar = covar_dict[last_covar_update_date]
                    w = pd.Series(weights).reindex(index=pd_covar.columns).fillna(0.0)
                    portfolio_vol[date] = np.sqrt(w.T @ pd_covar @ w)
        portfolio_vol = pd.Series(portfolio_vol, name=self.ticker)
        return portfolio_vol

    def compute_risk_contributions_implied_by_covar(self,
                                                    covar_dict: Dict[pd.Timestamp, pd.DataFrame] = None,
                                                    group_data: pd.Series = None,
                                                    group_order: List[str] = None,
                                                    freq: Optional[str] = None,
                                                    normalise: bool = False,
                                                    time_period: TimePeriod = None
                                                    ) -> pd.DataFrame:
        """
        compute risk contributions using covar_dict
        """
        if covar_dict is None:
            if self.covar_dict is None:
                raise ValueError(f"must provide covar_dict")
            else:
                covar_dict = self.covar_dict
        is_input_weights = True if freq is None else False
        strategy_weights = self.get_weights(freq=freq, is_input_weights=is_input_weights, time_period=time_period)
        covar_index = list(covar_dict.keys())
        strategy_rc = {}
        if freq is None:  # align with covar dates
            strategy_weights = strategy_weights.reindex(index=covar_index).ffill().fillna(0.0)
            for date, pd_covar in covar_dict.items():
                strategy_rc[date] = compute_portfolio_risk_contributions(w=strategy_weights.loc[date], covar=pd_covar)
        else:
            for date, weights in strategy_weights.to_dict(orient='index').items():
                if date > covar_index[0]:
                    last_covar_update_date = qis.find_upto_date_from_datetime_index(index=covar_index, date=date)
                    if last_covar_update_date is not None:
                        strategy_rc[date] = compute_portfolio_risk_contributions(w=pd.Series(weights).fillna(0.0),
                                                                             covar=covar_dict[last_covar_update_date])

        strategy_rc = pd.DataFrame.from_dict(strategy_rc, orient='index')
        if group_data is not None:
            strategy_rc = dfg.agg_df_by_groups_ax1(strategy_rc, group_data=group_data, group_order=group_order)
        if normalise:
            strategy_rc = qis.df_to_weight_allocation_sum1(strategy_rc)
        return strategy_rc

    # """
    # plotting methods
    # """

    def add_regime_shadows(self,
                           ax: plt.Subplot,
                           regime_benchmark: str,
                           index: pd.Index = None,
                           regime_classifier: BenchmarkReturnsQuantilesRegime = regime_classifier
                           ) -> None:
        """
        add regime shadows using regime_benchmark
        """
        if self.benchmark_prices is None:
            raise ValueError(f"set benchmarks data")
        pivot_prices = self.benchmark_prices[regime_benchmark]
        if index is not None:
            pivot_prices = pivot_prices.reindex(index=index, method='ffill')
        qis.add_bnb_regime_shadows(ax=ax, pivot_prices=pivot_prices, regime_classifier=regime_classifier)

    def plot_nav(self,
                 regime_benchmark: str = None,
                 time_period: TimePeriod = None,
                 add_benchmarks: bool = False,
                 regime_classifier: BenchmarkReturnsQuantilesRegime = regime_classifier,
                 ax: plt.Subplot = None,
                 **kwargs
                 ) -> None:
        if add_benchmarks:
            prices = self.get_portfolio_nav_with_benchmark_prices(time_period=time_period)
        else:
            prices = self.get_portfolio_nav(time_period=time_period)
        if ax is None:
            with sns.axes_style('darkgrid'):
                fig, ax = plt.subplots(1, 1, figsize=(16, 12), tight_layout=True)
        ppd.plot_prices(prices=prices, ax=ax, **kwargs)

        if regime_benchmark is not None:
            self.add_regime_shadows(ax=ax, regime_benchmark=regime_benchmark, index=prices.index,
                                    regime_classifier=regime_classifier)

    def plot_group_nav(self,
                       regime_benchmark: str = None,
                       time_period: TimePeriod = None,
                       add_benchmarks: bool = False,
                       regime_classifier: BenchmarkReturnsQuantilesRegime = regime_classifier,
                       ax: plt.Subplot = None,
                       **kwargs
                       ) -> None:

        total_group_navs = self.get_total_nav_with_group_navs(time_period=time_period)
        if add_benchmarks and self.benchmark_prices is not None:
            benchmark_prices = self.benchmark_prices.reindex(index=total_group_navs.index, method='ffill')
            total_group_navs = pd.concat([total_group_navs, benchmark_prices], axis=1)
        if ax is None:
            with sns.axes_style('darkgrid'):
                fig, ax = plt.subplots(1, 1, figsize=(16, 12), tight_layout=True)
        ppd.plot_prices(prices=total_group_navs, ax=ax, **kwargs)

        if regime_benchmark is not None:
            self.add_regime_shadows(ax=ax, regime_benchmark=regime_benchmark, index=total_group_navs.index,
                                    regime_classifier=regime_classifier)

    def plot_portfolio_vols(self,
                            time_period: TimePeriod = None,
                            freq: str = 'W-WED',
                            span: int = 13,  # 3m span of weekly returns
                            regime_benchmark: str = None,
                            regime_classifier: BenchmarkReturnsQuantilesRegime = regime_classifier,
                            ax: plt.Subplot = None,
                            **kwargs
                            ) -> plt.Figure:
        if ax is None:
            fig, ax = plt.subplots()

        portfolio_vols = self.compute_portfolio_vol(time_period=time_period, freq=freq, span=span)
        fig = qis.plot_time_series(df=portfolio_vols,
                                   var_format='{:,.2%}',
                                   legend_stats=qis.LegendStats.AVG_NONNAN_LAST,
                                   title=f"Portfolio EWM {span}-span vols of {freq}-freq returns",
                                   ax=ax,
                                   **kwargs)
        if regime_benchmark is not None:
            self.add_regime_shadows(ax=ax, regime_benchmark=regime_benchmark, index=portfolio_vols.index,
                                    regime_classifier=regime_classifier)
        return fig

    def plot_rolling_perf(self,
                          rolling_perf_stat: RollingPerfStat = RollingPerfStat.SHARPE,
                          add_benchmarks: bool = False,
                          regime_benchmark: str = None,
                          time_period: TimePeriod = None,
                          rolling_window: int = 1300,
                          roll_freq: Optional[str] = None,
                          legend_stats: pts.LegendStats = pts.LegendStats.AVG_LAST,
                          title: Optional[str] = None,
                          regime_classifier: BenchmarkReturnsQuantilesRegime = regime_classifier,
                          ax: plt.Subplot = None,
                          **kwargs
                          ) -> plt.Figure:

        # do not use start end dates here so the sharpe will be continuous with different time_period
        if add_benchmarks:
            prices = self.get_portfolio_nav_with_benchmark_prices(time_period=time_period)
        else:
            prices = self.get_portfolio_nav(time_period=time_period)
        if ax is None:
            fig, ax = plt.subplots()
        fig = ppd.plot_rolling_perf_stat(prices=prices,
                                         rolling_perf_stat=rolling_perf_stat,
                                         time_period=time_period,
                                         roll_periods=rolling_window,
                                         roll_freq=roll_freq,
                                         legend_stats=legend_stats,
                                         title=title,
                                         ax=ax,
                                         **kwargs)
        if regime_benchmark is not None:
            self.add_regime_shadows(ax=ax, regime_benchmark=regime_benchmark, index=prices.index,
                                    regime_classifier=regime_classifier)
        return fig

    def plot_ra_perf_table(self,
                           benchmark_price: Union[pd.Series, pd.DataFrame] = None,
                           benchmark: str = None,
                           is_grouped: bool = True,
                           time_period: TimePeriod = None,
                           perf_params: PerfParams = None,
                           perf_columns: List[PerfStat] = rpt.BENCHMARK_TABLE_COLUMNS,
                           title: str = None,
                           ax: plt.Subplot = None,
                           **kwargs
                           ) -> None:
        if is_grouped:
            total_nav = self.get_portfolio_nav(time_period=time_period)
            group_navs = self.get_group_navs(time_period=time_period, is_add_group_total=False)
            prices = pd.concat([total_nav, group_navs], axis=1)
            for_title = 'with portfolio groups'
            rows_edge_lines = [len(prices.columns)]
        else:
            prices = self.get_portfolio_nav(time_period=time_period).to_frame()
            for_title = ''
            rows_edge_lines = None
        if benchmark_price is not None:
            if isinstance(benchmark_price, pd.DataFrame):  # bechmark is the last asset
                if benchmark is None:
                    benchmark = benchmark_price.columns[-1]
            else:
                benchmark = benchmark_price.name
            if benchmark not in prices.columns:
                prices = pd.concat([prices, benchmark_price.reindex(index=prices.index, method='ffill')], axis=1)
            prices = prices.loc[:, ~prices.columns.duplicated(keep='first')]
            title = title or f"RA performance table {for_title} for {perf_params.freq_vol}-freq returns with beta to {benchmark}:" \
                             f" {qis.get_time_period(prices).to_str()}"
            ppt.plot_ra_perf_table_benchmark(prices=prices,
                                             benchmark=benchmark,
                                             perf_params=perf_params,
                                             perf_columns=perf_columns,
                                             title=title,
                                             rotation_for_columns_headers=0,
                                             special_rows_colors=[(1, 'deepskyblue'),
                                                                  (len(prices.columns), 'lightskyblue')],
                                             rows_edge_lines=rows_edge_lines,
                                             column_header='Portfolio',
                                             ax=ax,
                                             **kwargs)
        else:
            title = title or f"RA performance table {for_title}: {qis.get_time_period(prices).to_str()}"
            ppt.plot_ra_perf_table(prices=prices,
                                   perf_params=perf_params,
                                   perf_columns=rpt.COMPACT_TABLE_COLUMNS,
                                   title=title,
                                   rotation_for_columns_headers=0,
                                   column_header='Portfolio',
                                   ax=ax,
                                   **kwargs)

    def plot_returns_scatter(self,
                             benchmark_price: pd.Series,
                             is_grouped: bool = True,
                             time_period: TimePeriod = None,
                             title: str = None,
                             freq: str = 'QE',
                             ax: plt.Subplot = None,
                             **kwargs
                             ) -> None:
        if is_grouped:
            prices = self.get_total_nav_with_group_navs(time_period=time_period)
            title = title or f"Scatterplot of {freq}-freq returns by groups vs {str(benchmark_price.name)}"
        else:
            prices = self.get_portfolio_nav(time_period=time_period)
            title = title or f"Scatterplot of {freq}-freq returns vs {str(benchmark_price.name)}"
        prices = pd.concat([prices, benchmark_price.reindex(index=prices.index, method='ffill')], axis=1)
        if np.any(prices.columns.duplicated()):
            raise ValueError(f"duplicated columns {prices.columns}")
        local_kwargs = qis.update_kwargs(kwargs=kwargs,
                                         new_kwargs={'weight': 'bold',
                                                     'x_rotation': 0,
                                                     'first_color_fixed': False,
                                                     'ci': None,
                                                     'markersize': 6})
        prs.plot_returns_scatter(prices=prices,
                                 benchmark=str(benchmark_price.name),
                                 freq=freq,
                                 order=2,
                                 title=title,
                                 ax=ax,
                                 **local_kwargs)

    def plot_monthly_returns_heatmap(self,
                                     time_period: TimePeriod = None,
                                     ax: plt.Subplot = None,
                                     **kwargs
                                     ) -> None:
        # for monthly returns fix A and date_format
        kwargs = qis.update_kwargs(kwargs, dict(heatmap_freq='YE', date_format='%Y'))
        rhe.plot_returns_heatmap(prices=self.get_portfolio_nav(time_period=time_period),
                                 heatmap_column_freq='ME',
                                 is_add_annual_column=True,
                                 is_inverse_order=True,
                                 ax=ax,
                                 **kwargs)

    def plot_periodic_returns(self,
                              benchmark_prices: Union[pd.DataFrame, pd.Series] = None,
                              is_grouped: bool = True,
                              time_period: TimePeriod = None,
                              heatmap_freq: str = 'YE',
                              date_format: str = '%Y',
                              transpose: bool = True,
                              title: str = None,
                              ax: plt.Subplot = None,
                              **kwargs
                              ) -> None:
        if is_grouped:
            prices = self.get_total_nav_with_group_navs(time_period=time_period)
            title = title or f"{heatmap_freq}-returns by groups"
        else:
            prices = self.get_portfolio_nav(time_period=time_period).to_frame()
            title = title or f"{heatmap_freq}-returns"
        if benchmark_prices is not None:
            hline_rows = [1, len(prices.columns)]
            prices = pd.concat([prices, benchmark_prices.reindex(index=prices.index, method='ffill')], axis=1)
        else:
            hline_rows = None

        rhe.plot_periodic_returns_table(prices=prices,
                                        ax=ax,
                                        title=title,
                                        date_format=date_format,
                                        transpose=transpose,
                                        hline_rows=hline_rows,
                                        **qis.update_kwargs(kwargs, dict(freq=heatmap_freq)))

    def plot_regime_data(self,
                         benchmark_price: pd.Series,
                         is_grouped: bool = True,
                         regime_data_to_plot: RegimeData = RegimeData.REGIME_SHARPE,
                         time_period: TimePeriod = None,
                         var_format: Optional[str] = None,
                         is_conditional_sharpe: bool = True,
                         legend_loc: Optional[str] = 'upper center',
                         title: str = None,
                         perf_params: PerfParams = None,
                         regime_classifier: BenchmarkReturnsQuantilesRegime = BenchmarkReturnsQuantilesRegime(),
                         ax: plt.Subplot = None,
                         **kwargs
                         ) -> plt.Figure:

        if is_grouped:
            prices = self.get_total_nav_with_group_navs(time_period=time_period)
            title = title or (f"Sharpe ratio attribution by groups to {str(benchmark_price.name)} "
                              f"Bear/Normal/Bull regimes of {regime_classifier.freq} returns")
        else:
            prices = self.get_portfolio_nav(time_period=time_period).to_frame()
            title = title or (f"Sharpe ratio attribution to {str(benchmark_price.name)} Bear/Normal/Bull regimes "
                              f"of {regime_classifier.freq} returns")

        if benchmark_price.name not in prices.columns:
            prices = pd.concat([benchmark_price.reindex(index=prices.index, method='ffill'), prices], axis=1)

        fig = qis.plot_regime_data(regime_classifier=regime_classifier,
                                   prices=prices,
                                   benchmark=str(benchmark_price.name),
                                   is_conditional_sharpe=is_conditional_sharpe,
                                   regime_data_to_plot=regime_data_to_plot,
                                   var_format=var_format or '{:.2f}',
                                   legend_loc=legend_loc,
                                   perf_params=perf_params,
                                   title=title,
                                   ax=ax,
                                   **kwargs)
        return fig

    def plot_vol_regimes(self,
                         benchmark_price: pd.Series,
                         is_grouped: bool = True,
                         time_period: TimePeriod = None,
                         title: str = None,
                         freq: str = 'ME',
                         ax: plt.Subplot = None,
                         **kwargs
                         ) -> plt.Figure:

        if is_grouped:
            prices = self.get_total_nav_with_group_navs(time_period=time_period)
            title = title or f"{freq}-returns by groups conditional on vols {str(benchmark_price.name)}"
        else:
            prices = self.get_portfolio_nav(time_period=time_period)
            title = title or f"{freq}-returns conditional on vols {str(benchmark_price.name)}"
        prices = pd.concat([benchmark_price.reindex(index=prices.index, method='ffill'), prices], axis=1)

        fig = qis.plot_regime_boxplot(prices=prices,
                                      benchmark=str(benchmark_price.name),
                                      title=title,
                                      ax=ax,
                                      **kwargs)
        return fig

    def plot_contributors(self,
                          time_period: TimePeriod = None,
                          num_assets: int = 10,
                          ax: plt.Subplot = None,
                          **kwargs
                          ) -> None:
        prices = self.get_instruments_navs(time_period=time_period)
        ppt.plot_top_bottom_performers(prices=prices, num_assets=num_assets, ax=ax, **kwargs)

    def plot_best_worst_returns(self,
                                time_period: TimePeriod = None,
                                num_returns: int = 10,
                                ax: plt.Subplot = None,
                                **kwargs
                                ) -> None:
        price = self.get_portfolio_nav(time_period=time_period)
        ppt.plot_best_worst_returns(price=price, num_returns=num_returns, ax=ax, **kwargs)

    def plot_pnl(self, time_period: TimePeriod = None) -> None:
        avg_costs, realized_pnl, mtm_pnl, total_pnl, trades = self.compute_realized_pnl(time_period=time_period)
        prices = self.prices
        if time_period is not None:
            prices = time_period.locate(prices)
        with sns.axes_style('darkgrid'):
            fig, axs = plt.subplots(5, 1, figsize=(10, 16), tight_layout=True)
            pts.plot_time_series(df=prices, legend_stats=pts.LegendStats.FIRST_AVG_LAST, title='prices', ax=axs[0])
            pts.plot_time_series(df=avg_costs, legend_stats=pts.LegendStats.FIRST_AVG_LAST, title='avg_costs',
                                 ax=axs[1])
            pts.plot_time_series(df=realized_pnl, legend_stats=pts.LegendStats.FIRST_AVG_LAST, title='realized_pnl',
                                 ax=axs[2])
            pts.plot_time_series(df=mtm_pnl, legend_stats=pts.LegendStats.FIRST_AVG_LAST, title='mtm_pnl', ax=axs[3])
            pts.plot_time_series(df=total_pnl, legend_stats=pts.LegendStats.FIRST_AVG_LAST, title='total_pnl',
                                 ax=axs[4])

    def plot_weights(self,
                     is_input_weights: bool = True,
                     add_mean_levels: bool = False,
                     use_bar_plot: bool = False,
                     columns: List[str] = None,
                     freq: Optional[str] = None,
                     is_yaxis_limit_01: bool = True,
                     bbox_to_anchor: Tuple[float, float] = None,
                     legend_stats: pst.LegendStats = pst.LegendStats.FIRST_AVG_LAST,
                     var_format: str = '{:.1%}',
                     title: Optional[str] = None,
                     ax: plt.Subplot = None,
                     **kwargs
                     ) -> None:
        weights = self.get_weights(is_input_weights=is_input_weights,
                                   columns=columns,
                                   freq=freq)
        pst.plot_stack(df=weights,
                       add_mean_levels=add_mean_levels,
                       use_bar_plot=use_bar_plot,
                       is_yaxis_limit_01=is_yaxis_limit_01,
                       bbox_to_anchor=bbox_to_anchor,
                       title=title,
                       legend_stats=legend_stats,
                       var_format=var_format,
                       ax=ax,
                       **kwargs)

    def plot_performance_attribution(self,
                                     time_period: TimePeriod = None,
                                     attribution_metric: AttributionMetric = AttributionMetric.PNL,
                                     add_top_bar_values: Optional[bool] = None,
                                     remove_zero_data: bool = True,
                                     shorten_instrument_names: bool = True,
                                     ax: plt.Subplot = None,
                                     **kwargs
                                     ) -> None:
        """
        performance attribution for p&l and risk
        shorten_instrument_names: keep 15 strings to improve axis alignment
        """
        data = self.get_performance_attribution_data(attribution_metric=attribution_metric,
                                                     time_period=time_period,
                                                     **kwargs)
        if remove_zero_data:
            # data = data.replace({0.0: np.nan}).dropna()
            data[np.isclose(data, 0.0)] = np.nan
            data = data.dropna()

        if shorten_instrument_names:
            data.index = [x[:15] for x in data.index]

        if add_top_bar_values is None:
            if isinstance(data, pd.Series) and len(data.index) <= 20:
                add_top_bar_values = True
            else:
                add_top_bar_values = False

        kwargs = qis.update_kwargs(kwargs=kwargs,
                                   new_kwargs=dict(bbox_to_anchor=(0.5, 1.05),
                                                   add_top_bar_values=add_top_bar_values,
                                                   x_rotation=90))
        title = f"{attribution_metric.value}"
        if time_period is not None:
            title += f", for period={time_period.to_str()}"
        qis.plot_bars(df=data,
                      skip_y_axis=True,
                      title=title,
                      stacked=False,
                      yvar_format='{:,.1%}' if add_top_bar_values else '{:,.2%}',
                      ax=ax,
                      **kwargs)

    def plot_benchmark_betas(self,
                             benchmark_prices: pd.DataFrame,
                             regime_benchmark: str = None,
                             time_period: TimePeriod = None,
                             freq: str = 'W-WED',
                             title: str = None,
                             beta_span: int = 52,
                             regime_classifier: BenchmarkReturnsQuantilesRegime = None,
                             add_zero_line: bool = True,
                             ax: plt.Subplot = None,
                             **kwargs
                             ) -> None:
        factor_exposures = self.compute_portfolio_benchmark_betas(benchmark_prices=benchmark_prices,
                                                                  time_period=time_period,
                                                                  freq_beta=freq,
                                                                  factor_beta_span=beta_span)
        qis.plot_time_series(df=factor_exposures,
                             var_format='{:,.2f}',
                             legend_stats=qis.LegendStats.AVG_NONNAN_LAST,
                             title=title or f"Portfolio rolling {beta_span}-span Betas to Benchmarks",
                             ax=ax,
                             **kwargs)
        if regime_benchmark is not None:
            self.add_regime_shadows(ax=ax, regime_benchmark=regime_benchmark, index=factor_exposures.index,
                                    regime_classifier=regime_classifier)

        if add_zero_line:
            ax.axhline(0, color='black', lw=1)

    def plot_portfolio_grouped_var(self,
                                   is_correlated: bool = True,
                                   regime_benchmark: str = None,
                                   time_period: TimePeriod = None,
                                   freq: str = 'B',
                                   title: str = None,
                                   total_column: Optional[str] = 'Total',
                                   vol_span: int = 33,  # span in number of freq-returns
                                   regime_classifier: BenchmarkReturnsQuantilesRegime = None,
                                   ax: plt.Subplot = None,
                                   **kwargs
                                   ) -> None:
        portfolio_vars, instrument_vars = self.compute_portfolio_vars(is_correlated=is_correlated,
                                                                      time_period=time_period,
                                                                      freq=freq,
                                                                      total_column=total_column,
                                                                      vol_span=vol_span)
        if is_correlated:
            title = title or f"Correlated {freq}-freq 99%-VAR with {vol_span}-span ewma covar"
        else:
            title = title or f"Independent {freq}-freq 99%-VAR with {vol_span}-span ewma vols"

        qis.plot_time_series(df=portfolio_vars,
                             var_format='{:,.2%}',
                             legend_stats=qis.LegendStats.AVG_MIN_MAX_LAST,
                             title=title,
                             y_limits=(0.0, None),
                             ax=ax,
                             **kwargs)

        if regime_benchmark is not None:
            self.add_regime_shadows(ax=ax, regime_benchmark=regime_benchmark, index=portfolio_vars.index,
                                    regime_classifier=regime_classifier)

    def plot_current_weights(self,
                             is_grouped: bool = False,
                             add_top_bar_values: Optional[bool] = None,
                             time_period: TimePeriod = None,
                             title: str = None,
                             group_data: pd.Series = None,
                             group_order: List[str] = None,
                             ax: plt.Subplot = None,
                             **kwargs
                             ) -> None:
        weights = self.get_weights(is_input_weights=True, freq=None, is_grouped=is_grouped,
                                   group_data=group_data, group_order=group_order)
        if time_period is not None:
            weights_1 = time_period.locate(weights)
            if len(weights_1.index) > 0:
                weights_1 = weights_1.iloc[0, :]
            else:
                weights_1 = weights.iloc[-1, :]
        else:
            weights_1 = weights.iloc[-1, :]
        if add_top_bar_values is None:
            if len(weights_1.index) <= 10:
                add_top_bar_values = True
            else:
                add_top_bar_values = False
        kwargs = qis.update_kwargs(kwargs=kwargs, new_kwargs=dict(legend_loc=None,
                                                                  add_top_bar_values=add_top_bar_values,
                                                                  x_rotation=90))
        if title is None:
            if is_grouped:
                title = f"Weights by groups on {qis.date_to_str(weights_1.name)}"
            else:
                title = f"Weights by instruments on {qis.date_to_str(weights_1.name)}"
        qis.plot_bars(df=weights_1,
                      skip_y_axis=True,
                      title=title,
                      stacked=False,
                      yvar_format='{:,.2%}' if add_top_bar_values else '{:,.2%}',
                      ax=ax,
                      **kwargs)

    def plot_last_weights_change(self,
                                 is_grouped: bool = False,
                                 title: str = None,
                                 add_top_bar_values: Optional[bool] = None,
                                 weights_change_lag: Optional[int] = None,
                                 ax: plt.Subplot = None,
                                 **kwargs
                                 ) -> None:
        weights = self.get_weights(is_input_weights=True, freq=None, is_grouped=is_grouped)
        weights_1 = weights.iloc[-1, :]
        if len(weights.index) > 1:
            if weights_change_lag is not None:
                weights_0 = weights.iloc[-weights_change_lag, :]
            else:
                weights_0 = weights.iloc[-2, :]
            delta = weights_1.subtract(weights_0)
            post_title = f"rebalancing between {qis.date_to_str(weights_0.name)} and {qis.date_to_str(weights_1.name)}"
        else:
            delta = weights_1
            post_title = f"rebalancing on {qis.date_to_str(weights_1.name)} starting from zero"
        if add_top_bar_values is None:
            if len(delta.index) <= 10:
                add_top_bar_values = True
            else:
                add_top_bar_values = False
        kwargs = qis.update_kwargs(kwargs=kwargs, new_kwargs=dict(legend_loc=None,
                                                                  add_top_bar_values=add_top_bar_values,
                                                                  x_rotation=90))
        if title is None:
            if is_grouped:
                title = f"Change in weights by groups with {post_title}"
            else:
                title = f"Change in weights by instruments with {post_title}"
        qis.plot_bars(df=delta,
                      skip_y_axis=True,
                      title=title,
                      stacked=False,
                      yvar_format='{:,.2%}' if add_top_bar_values else '{:,.2%}',
                      ax=ax,
                      **kwargs)

    def plot_current_var(self,
                         snapshot_period: SnapshotPeriod = SnapshotPeriod.LAST,
                         is_grouped: bool = False,
                         is_correlated: bool = True,
                         time_period: TimePeriod = None,
                         freq: str = 'B',
                         title: str = None,
                         add_top_bar_values: Optional[bool] = None,
                         total_column: Optional[str] = 'Total',
                         vol_span: Union[int, float] = 33,  # span in number of freq-returns
                         ax: plt.Subplot = None,
                         **kwargs
                         ) -> None:
        portfolio_vars, instrument_vars = self.compute_portfolio_vars(is_correlated=is_correlated,
                                                                      time_period=time_period,
                                                                      freq=freq,
                                                                      total_column=total_column,
                                                                      vol_span=vol_span)
        if is_grouped:
            if snapshot_period == SnapshotPeriod.LAST:
                var_1 = portfolio_vars.iloc[-1, :]
                if is_correlated:
                    title = title or f"Correlated {freq}-freq 99%-VAR with {vol_span}-span ewma covar: {qis.date_to_str(var_1.name)}"
                else:
                    title = title or f"Independent {freq}-freq 99%-VAR with {vol_span}-span ewma vols: {qis.date_to_str(var_1.name)}"
            elif snapshot_period == SnapshotPeriod.AVG:
                var_1 = portfolio_vars.mean(0)
                period = qis.get_time_period(df=portfolio_vars).to_str(date_separator='-')
                if is_correlated:
                    title = title or f"Avg Correlated {freq}-freq 99%-VAR with {vol_span}-span: {period}"
                else:
                    title = title or f"Avg Independent {freq}-freq 99%-VAR with {vol_span}-span: {period}"
            elif snapshot_period == SnapshotPeriod.MAX:
                var_1 = portfolio_vars.max(0)
                period = qis.get_time_period(df=portfolio_vars).to_str(date_separator='-')
                if is_correlated:
                    title = title or f"Max Correlated {freq}-freq 99%-VAR with {vol_span}-span: {period}"
                else:
                    title = title or f"Max Independent {freq}-freq 99%-VAR with {vol_span}-span: {period}"
            else:
                raise NotImplementedError(f"snapshot_period={snapshot_period}")

        else:
            if is_correlated:
                raise ValueError(f"instrument var is not defined for correlated var ")
            else:
                if snapshot_period == SnapshotPeriod.LAST:
                    var_1 = instrument_vars.iloc[-1, :]
                    title = title or f"Instrument independent {freq}-freq 99%-VAR with {vol_span}-span ewma vols: {qis.date_to_str(var_1.name)}"
                elif snapshot_period == SnapshotPeriod.AVG:
                    var_1 = instrument_vars.mean(0)
                    period = qis.get_time_period(df=instrument_vars).to_str()
                    title = title or f"Avg Independent {freq}-freq 99%-VAR with {vol_span}-span ewma vols: {period}"
                elif snapshot_period == SnapshotPeriod.MAX:
                    var_1 = instrument_vars.max(0)
                    period = qis.get_time_period(df=instrument_vars).to_str()
                    title = title or f"Max Independent {freq}-freq 99%-VAR with {vol_span}-span ewma vols: {period}"
                else:
                    raise NotImplementedError(f"snapshot_period={snapshot_period}")

        if add_top_bar_values is None:
            if len(var_1.index) <= 10:
                add_top_bar_values = True
            else:
                add_top_bar_values = False
        kwargs = qis.update_kwargs(kwargs=kwargs, new_kwargs=dict(legend_loc=None,
                                                                  add_bar_values=add_top_bar_values,
                                                                  x_rotation=90))
        qis.plot_bars(df=var_1,
                      skip_y_axis=True,
                      title=title,
                      stacked=False,
                      yvar_format='{:,.2%}',
                      ax=ax,
                      **kwargs)

    def plot_var_stack(self,
                       is_grouped: bool = True,
                       is_correlated: bool = True,
                       time_period: TimePeriod = None,
                       freq: str = 'B',
                       title: str = None,
                       stack_freq: Optional[str] = 'W-WED',
                       vol_span: int = 33,  # span in number of freq-returns
                       ax: plt.Subplot = None,
                       **kwargs
                       ) -> None:
        portfolio_vars, instrument_vars = self.compute_portfolio_vars(is_correlated=is_correlated,
                                                                      time_period=time_period,
                                                                      freq=freq,
                                                                      total_column=None,
                                                                      vol_span=vol_span)
        if is_grouped:
            var_1 = portfolio_vars
            if is_correlated:
                title = title or f"Risk Attribution with Correlated {freq}-freq 99%-VAR with {vol_span}-span ewma covar"
            else:
                title = title or f"Risk Attribution with Independent {freq}-freq 99%-VAR with {vol_span}-span ewma vols "

        else:
            if is_correlated:
                raise ValueError(f"instrument var is not defined for correlated var ")
            else:
                var_1 = instrument_vars
                title = title or f"Risk Attribution with Instrument independent {freq}-freq 99%-VAR with {vol_span}-span ewma vols"

        if time_period is not None:
            var_1 = time_period.locate(var_1)
        if stack_freq is not None:
            var_1 = var_1.asfreq(stack_freq, method='ffill')
        df = var_1.divide(np.nansum(var_1, axis=1, keepdims=True))
        qis.plot_stack(df=df,
                       is_yaxis_limit_01=True,
                       use_bar_plot=True,
                       title=title,
                       legend_stats=qis.LegendStats.AVG_NONNAN_LAST,
                       var_format='{:.1%}',
                       ax=ax,
                       **qis.update_kwargs(kwargs, dict(bbox_to_anchor=(0.5, 1.05), ncols=2)))

    def plot_distribution_yield(self,
                                paid_dividends: pd.DataFrame,
                                time_period: TimePeriod = None,
                                div_rolling_freq: str = 'ME',
                                div_rolling_period: int = 12,
                                axs: List[plt.Subplot] = None,
                                **kwargs
                                ):
        distributions_by_instrument_yield_12m, distribution_yield, distribution_yield_12m = \
            self.compute_distribution_yield(paid_dividends=paid_dividends,
                                            time_period=time_period,
                                            div_rolling_freq=div_rolling_freq,
                                            div_rolling_period=div_rolling_period)
        if axs is None:
            fig1, ax1 = plt.subplots(1, 1, figsize=(14, 10), constrained_layout=True)
            fig2, ax2 = plt.subplots(1, 1, figsize=(14, 10), constrained_layout=True)
            axs = [ax1, ax2]

        qis.plot_time_series(distribution_yield_12m,
                             ax=axs[0],
                             **kwargs)
        qis.plot_stack(df=distributions_by_instrument_yield_12m.resample('ME').last(),
                       use_bar_plot=True,
                       ax=axs[1],
                       **kwargs)

    def plot_turnover(self,
                      regime_benchmark: str = None,
                      is_agg: bool = False,
                      is_grouped: bool = False,
                      group_data: pd.Series = None,
                      group_order: List[str] = None,
                      time_period: TimePeriod = None,
                      turnover_rolling_period: Optional[int] = 260,
                      freq_turnover: Optional[str] = 'B',
                      add_total: bool = True,
                      title: str = None,
                      regime_classifier: BenchmarkReturnsQuantilesRegime = None,
                      ax: plt.Subplot = None,
                      **kwargs
                      ) -> None:
        turnover = self.get_turnover(is_agg=is_agg,
                                     is_grouped=is_grouped,
                                     group_data=group_data,
                                     group_order=group_order,
                                     time_period=time_period,
                                     roll_period=turnover_rolling_period,
                                     add_total=add_total,
                                     freq=freq_turnover,
                                     **kwargs)
        if not turnover.empty:
            freq = pd.infer_freq(turnover.index)
            turnover_title = title or f"{turnover_rolling_period}-period rolling {freq}-freq Turnover"
            qis.plot_time_series(df=turnover,
                                 var_format='{:,.1%}',
                                 y_limits=(0.0, None),
                                 legend_stats=qis.LegendStats.AVG_NONNAN_LAST,
                                 title=turnover_title,
                                 ax=ax,
                                 **kwargs)
            if regime_benchmark is not None and self.benchmark_prices is not None:
                self.add_regime_shadows(ax=ax, regime_benchmark=regime_benchmark, index=turnover.index,
                                        regime_classifier=regime_classifier)


@njit
def compute_realized_pnl(prices: np.ndarray,
                         units: np.ndarray
                         ) -> Tuple[np.ndarray, ...]:
    """
    pnl for long only positions, computes avg entry price and pnl by instrument
    """
    avg_costs = np.zeros_like(prices)
    realized_pnl = np.zeros_like(prices)
    mtm_pnl = np.zeros_like(prices)
    trades = np.zeros_like(prices)
    for idx, (price1, unit1) in enumerate(zip(prices, units)):
        if idx == 0:
            avg_costs[idx] = np.where(np.greater(unit1, 1e-16), price1, 0.0)
        else:
            unit0 = units[idx - 1]
            avg_costs0 = avg_costs[idx - 1]
            delta = unit1 - unit0
            is_purchase = np.greater(delta, 1e-16)
            is_sell = np.less(delta, -1e-16)
            realized_pnl[idx] = np.where(is_sell, -delta * (price1 - avg_costs0), 0.0)
            avg_costs[idx] = np.where(is_purchase, np.true_divide(delta * price1 + unit0 * avg_costs0, unit1),
                                      avg_costs0)
            mtm_pnl[idx] = unit0 * (price1 - avg_costs0) - realized_pnl[idx]
            trades[idx] = delta
    return avg_costs, realized_pnl, mtm_pnl, trades


class AllocationType(EnumMap):
    EW = 1
    FIXED_WEIGHTS = 2
    ERC = 3
    ERC_ALT = 4


@dataclass
class PortfolioInput:
    """
    define data inputs for portfolio construction
    """
    name: str
    weights: Union[np.ndarray, pd.DataFrame, Dict[str, float]]
    prices: pd.DataFrame = None  # mandatory but we set none for enumarators
    allocation_type: AllocationType = AllocationType.FIXED_WEIGHTS
    time_period: TimePeriod = None
    rebalancing_freq: str = 'QE'
    freq_regime: str = 'ME'
    returns_freq: str = 'ME'
    ewm_lambda: float = 0.92
    target_vol: float = None

    def update(self, new: Dict[str, Any]):
        for key, value in new.items():
            if hasattr(self, key):
                setattr(self, key, value)




# packages
import matplotlib.pyplot as plt
import seaborn as sns
from enum import Enum

from qis.perfstats.regime_classifier import (RegimeClassifier,
                                             BenchmarkReturnsQuantilesRegime,
                                             BenchmarkReturnsQuantilesRegime,
                                             BenchmarkReturnsPositiveNegativeRegime)
from qis.plots.derived.regime_scatter import plot_scatter_regression


class LocalTests(Enum):
    """Enumeration of available local test cases."""
    SCATTER_ALL = 1
    SCATTER_1 = 2


def run_local_test(local_test: LocalTests):
    """
    Run local tests for development and debugging purposes.

    Integration tests that download real data and generate regression plots.
    Use for quick verification during development.

    Args:
        local_test: Which test case to run
    """
    from qis.test_data import load_etf_data

    # Load sample ETF price data
    prices = load_etf_data().dropna()
    # regime_classifier = BenchmarkReturnsQuantilesRegime(freq='QE')
    regime_classifier = BenchmarkReturnsPositiveNegativeRegime()

    if local_test == LocalTests.SCATTER_ALL:
        # Test with all assets
        with sns.axes_style('darkgrid'):
            plot_scatter_regression(
                prices=prices,
                regime_benchmark='SPY',
                regime_classifier=regime_classifier,
                add_last_date=False,
                display_estimated_betas=True
            )

    elif local_test == LocalTests.SCATTER_1:
        # Test with single asset pair
        with sns.axes_style('darkgrid'):
            plot_scatter_regression(
                prices=prices[['SPY', 'TLT']],
                regime_benchmark='SPY',
                regime_classifier=regime_classifier,
                add_last_date=True,
                display_estimated_betas=True
            )

    plt.show()


if __name__ == '__main__':
    run_local_test(local_test=LocalTests.SCATTER_ALL)




# packages
import numpy as np
import pandas as pd
from dataclasses import dataclass
from enum import Enum
from typing import NamedTuple, Optional, Callable, Dict, List

# qis
import qis.file_utils as fu
import qis.utils.df_groups as dfg
import qis.utils.df_agg as dfa
import qis.utils.struct_ops as sop
import qis.utils.df_str as dff

DATE_FORMAT = '%d%b%y' # short y for meta


class ValueType(Enum):
    FLOAT0 = 0
    FLOAT = 1
    FLOAT2 = 2
    FLOAT4 = 14
    COMMA_FLOAT0 = 3
    PRICE = 4
    PERCT = 5
    PERCT0 = 6
    DATE = 7
    INT = 8
    STR = 9
    WEIGHT = 10
    SHARPE = 11
    MARG_SHARPE = 12


class ColVar(NamedTuple):
    """
    variable characteristic for columns variable for table
    """
    name: str
    short: str = None
    short_n: str = None
    value_type: ValueType = ValueType.FLOAT
    agg_func: Optional[Callable] = None

    def to_str(self, short: bool = False, short_n: bool = False, **kwargs) -> str:
        if short and self.short is not None:
            name = self.short
        elif short_n and self.short_n is not None:
            name = self.short_n
        else:
            name = self.name
        return name

    def to_format(self,
                  digits_to_show: int = 2,
                  sharpe_digits: int = 2,
                  price_format: Optional[str] = None,
                  # date_format: str = DATE_FORMAT,
                  **kwargs
                  ) -> str:

        var_format = '{:.2f}'

        if self.value_type == ValueType.FLOAT:
            if digits_to_show == 1:
                var_format = '{:.1f}'
            elif digits_to_show == 4:
                var_format = '{:.4f}'
            else:
                var_format = '{:.2f}'
        elif self.value_type == ValueType.FLOAT0:
            var_format = '{:.0f}'
        elif self.value_type == ValueType.FLOAT2:
            var_format = '{:.2f}'
        elif self.value_type == ValueType.FLOAT4:
            var_format = '{:.4f}'
        elif self.value_type == ValueType.COMMA_FLOAT0:
            var_format = '{:,.0f}'
        elif self.value_type == ValueType.PRICE:
            if price_format is not None:
                var_format = price_format
            else:
                var_format = '{:.2f}'

        elif self.value_type == ValueType.PERCT:
            if digits_to_show == 0:
                var_format = '{:.0%}'
            elif digits_to_show == 1:
                var_format = '{:.1%}'
            else:
                var_format = '{:.2%}'

        elif self.value_type == ValueType.PERCT0:
            var_format = '{:,.0%}'

        elif self.value_type == ValueType.DATE:
             var_format = DATE_FORMAT  # independent

        elif self.value_type == ValueType.INT:
             var_format = '{:.0f}'

        elif self.value_type == ValueType.STR:
             var_format = '{}'

        elif self.value_type == ValueType.WEIGHT:
             var_format = '{:.2%}'

        elif self.value_type == ValueType.SHARPE:
            if sharpe_digits == 2:
                var_format = '{:.2f}'
            else:
                var_format = '{:.1f}'

        elif self.value_type == ValueType.MARG_SHARPE:
            var_format = '{:.3f}'

        else:
            TypeError(f"unsupported type = {self.value_type}")

        return var_format


@dataclass
class ColumnData:
    """
    keep data for producing meta and aggregation
    """
    column: ColVar
    data: pd.Series
    is_total_mean: bool = True


def column_datas_to_df(column_datas: Dict[str, ColumnData],
                       weight: Optional[pd.Series] = None,
                       agg_column: Optional[str] = None,
                       group_order: Optional[List[str]] = None,
                       sort_column: str = None,
                       ascending: bool = False,
                       is_to_str: bool = True,
                       total_name: str = 'Total',
                       new_index_col: str = None,
                       agg_columns_filling: Optional[Dict[str, str]] = {'Name': 'Avg'}
                       ) -> pd.DataFrame:
    """
    aggregate dictionary of column vars to pandas with aggregation
    """
    datas = []
    for _, column_data in column_datas.items():
        datas.append(column_data.data.rename(column_data.column.to_str()))
    table_data = pd.concat(datas, axis=1)

    if weight is None:
        weight = pd.Series(1.0, index=table_data.index)

    # now sort
    if sort_column is not None and agg_column is not None:
        table_data = dfg.sort_df_by_index_group(df=table_data,
                                                group_column=agg_column,
                                                sort_column=sort_column,
                                                group_order=group_order,
                                                ascending=ascending)

    if new_index_col is not None:
        table_data = table_data.set_index(new_index_col, drop=True)
        column_datas.pop(new_index_col)

    # create dict for ac aggregation
    if agg_column is not None:
        agg_dict = {}
        table_data1 = table_data.copy()
        for _, column_data in column_datas.items():
            if column_data.column.agg_func is not None:
                if column_data.column.agg_func == dfa.sum_weighted:
                    table_data1[column_data.column.to_str()] = table_data1[column_data.column.to_str()].multiply(weight)
                    agg_dict[column_data.column.to_str()] = np.nansum
                else:
                    agg_dict[column_data.column.to_str()] = column_data.column.agg_func

        if len(agg_dict.keys()) > 0:
            try:
                agg_table = table_data1.groupby(agg_column).agg(agg_dict)  # use agg as apply func for columns
            except AttributeError:
                raise AttributeError(f"dublicated columnes\n{agg_column}")

            # arrange by ac order add name
            if group_order is not None and agg_table is not None:
                groups = sop.merge_lists_unique(list1=group_order, list2=list(np.unique(agg_table.index)))
                groups1 = [x for x in groups if x in agg_table.index]
                agg_table = agg_table.loc[groups1, :]

            # add total as means
            port_table = agg_table.mean(axis=0).to_frame(total_name).T
            # replace with sum
            for _, column_data in column_datas.items():
                if column_data.is_total_mean is False:
                    port_table[column_data.column.to_str()] = np.nansum(agg_table[column_data.column.to_str()])

            agg_table_data = pd.concat([table_data, agg_table, port_table], axis=0)

            # for asset class data in column name add avg for clarity
            if agg_columns_filling is not None:
                for col, value in agg_columns_filling.items():
                    if col in agg_table_data.columns:
                        agg_table_data.loc[agg_table.index, col] = value
                        agg_table_data.loc[port_table.index, col] = value

        else:
            print("no vars for aggregation")
            agg_table_data = table_data
    else:
        agg_table_data = table_data

    if is_to_str:  # convert to str:
        for _, column_data in column_datas.items():
            agg_table_data[column_data.column.to_str()] = dff.series_to_str(
                ds=agg_table_data[column_data.column.to_str()],
                var_format=column_data.column.to_format(digits_to_show=2))
    return agg_table_data


@dataclass
class DfOutDict:
    df_out_dict: Dict[str, pd.DataFrame] = None
    last_data: Optional[pd.DataFrame] = None

    def __post_init__(self):
        self.df_out_dict = {}

    def __getitem__(self, key: str) -> pd.DataFrame:
        # for df_out_dict[key]
        return self.df_out_dict[key]

    def append(self, df: pd.DataFrame, name: str) -> None:
        if name not in self.df_out_dict.keys():
            self.df_out_dict[name] = df
        else:
            raise ValueError(f"{name} exist in {self.df_out_dict.keys()}")

    def set_last_df(self, data: pd.DataFrame) -> None:
        # way to first make a placement and then append it
        self.last_data = data

    def append_last_df(self, name: str) -> None:
        # way to first make a placement and then append it
        if self.last_data is None:
            raise ValueError(f"self.last_data is None")
        self.append(df=self.last_data, name=name)
        self.last_data = None

    def print(self):
        for key, df in self.df_out_dict.items():
            print(f"{key}")
            print(df)

    def save(self, file_name: str) -> None:
        if len(self.df_out_dict.keys()) > 0:
            file_path = fu.save_df_to_excel(data=self.df_out_dict, file_name=file_name)
            print(f"saved output data to excel:\n {file_path}")


class EnumMap(Enum):
    """
    abstract enum with a map function
    """
    @classmethod
    def map_to_value(cls, name):
        """
        given name return value
        """
        for k, v in cls.__members__.items():
            if k == name:
                return v
        raise ValueError(f"nit in enum {name}")


class DotDict(dict):
    """
    dot.notation access to dictionary attributes
    """
    def __getattr__(self, attr):
        return self.get(attr)
    __setattr__ = dict.__setitem__
    __delattr__ = dict.__delitem__

    def __getstate__(self):
        return self

    def __setstate__(self, state):
        self.update(state)
        self.__dict__ = self

    # def __setattr__(self, name, value):
    #    self._data[name] = value


class LocalTests(Enum):
    DOT_DICT = 1


def run_local_test(local_test: LocalTests):
    """Run local tests for development and debugging purposes.

    These are integration tests that download real data and generate reports.
    Use for quick verification during development.
    """

    if local_test == LocalTests.DOT_DICT:
        this = DotDict({'me': 3, 'you': 10})
        print(this)
        print(this.me)
        print(this.you)

        for k, v in this.items():
            print(f"{k}: {v}")

        this['me1'] = 6
        this.me2 = 12
        for k, v in this.items():
            print(f"{k}: {v}")


if __name__ == '__main__':

    run_local_test(local_test=LocalTests.DOT_DICT)



# packages
import numpy as np
import pandas as pd
import matplotlib as mpl
import matplotlib.pyplot as plt
import seaborn as sns
from typing import List, Optional, Tuple
from enum import Enum

# qis
import qis.plots.utils as put
from qis.plots.utils import LegendStats


def plot_stack(df: pd.DataFrame,
               use_bar_plot: bool = False,
               is_yaxis_limit_01: bool = False,
               add_mean_levels: bool = False,
               add_cum_levels: bool = False,
               add_total_line: bool = False,
               colors: List[str] = None,
               step: Optional[str] = None,  # 'mid
               title: Optional[str] = None,
               baseline: str = 'zero',  # "zero", "sym", "wiggle", "weighted_wiggle"
               ncols: int = 1,
               legend_loc: Optional[str] = 'upper center',
               legend_labels: Optional[List[str]] = None,
               legend_stats: LegendStats = LegendStats.NONE,
               var_format: str = '{:.0%}',
               fontsize: int = 10,
               linewidth: float = 1.5,
               x_rotation: int = 90,
               reverse_columns: bool = False,
               x_date_freq: str = 'YE',
               date_format: str = '%b-%y',
               skip_y_axis: bool = True,
               bbox_to_anchor: Optional[Tuple[float, float]] = None,
               xlabel: str = None,
               ylabel: str = None,
               ax: plt.Subplot = None,
               **kwargs
               ) -> plt.Figure:

    if ax is None:
        fig, ax = plt.subplots()
    else:
        fig = None

    if isinstance(df.index, pd.DatetimeIndex):
        re_indexed_data, datalables = put.map_dates_index_to_str(data=df,
                                                                 x_date_freq=x_date_freq,
                                                                 date_format=date_format)
    else:
        re_indexed_data = df
        datalables = None

    if colors is None:
        colors = put.get_n_colors(n=len(re_indexed_data.columns))

    if use_bar_plot:  # plot bar apperas to look better for unconstraint plots
        re_indexed_data.plot.bar(stacked=True,
                                 color=colors, width=1.0, alpha=1.0, edgecolor='none',
                                 linewidth=0, ax=ax)
    else:
        ax.stackplot(re_indexed_data.index, re_indexed_data.T,
                     labels=re_indexed_data.columns, step=step, colors=colors,
                     baseline=baseline, edgecolor='none')

    # set x axes to nearest years
    ax.set_xlim(re_indexed_data.index[0], re_indexed_data.index[-1])

    if add_total_line:  # add total as line
        totals = re_indexed_data.sum(1)
        sns.lineplot(x=re_indexed_data.index, y=totals, marker='None', color='black', ax=ax)
        # legend_labels.append('Total')
        colors.append('black')

    # change axes labels, positions of each tick, relative to the indices of the x-values
    if datalables is not None:
        current_ticks = ax.get_xticks()
        ax.set_xticks(np.linspace(current_ticks[0], current_ticks[-1], len(datalables)))
        ax.set_xticklabels(datalables, rotation=90, fontsize=fontsize)

    if is_yaxis_limit_01:
        ax.set_ylim(0, 1)

    if add_mean_levels or add_cum_levels:
        xmin, xmax = ax.get_xlim()
        ymin, ymax = ax.get_ylim()
        cum_mean = 0.0
        cum_mean0 = 0.0
        handles, labels = ax.get_legend_handles_labels()

        for (idx, column), handle, label in zip(enumerate(re_indexed_data.columns), handles, labels):
            mean = np.mean(re_indexed_data[column].values)
            cum_mean = cum_mean + mean

            if add_mean_levels:
                vlabel = var_format.format(mean)
            else:
                vlabel = var_format.format(cum_mean)

            if add_mean_levels and is_yaxis_limit_01 is False:  # show absolute effect
                y_loc = mean
            else:  # show cumulative effect
                if column == re_indexed_data.columns[-1]:  # make it vidsiblae
                    y_loc = 1.0 - cum_mean0 if cum_mean0 < 0.5 else cum_mean0 + 0.5 * mean
                else:
                    y_loc = cum_mean
            cum_mean0 = cum_mean

            ax.axhline(y_loc, color='black', linestyle='--', linewidth=linewidth)

            color = mpl.colors.to_rgb(handle.get_facecolors()[0])
            ax.annotate(text=f"{label}={vlabel}", xy=(xmax, y_loc), fontsize=fontsize, weight='normal', color=color)

        y_annotation = 'Avg' if add_mean_levels else 'Total'
        ax.annotate(y_annotation, xy=(xmax, ymax), xytext=(1, 2), fontsize=fontsize, weight='normal',
                    textcoords='offset points', ha='left', va='bottom')

    put.set_ax_ticks_format(ax=ax, fontsize=fontsize, xvar_format=None, yvar_format=var_format)
    put.set_ax_tick_labels(ax=ax, x_rotation=x_rotation, fontsize=fontsize, skip_y_axis=skip_y_axis, **kwargs)
    put.set_ax_xy_labels(ax=ax, xlabel=xlabel, ylabel=ylabel, **kwargs)

    if legend_loc is not None:
        legend_title = None
        if legend_labels is None:
            legend_labels = put.get_legend_lines(data=re_indexed_data,
                                                 legend_stats=legend_stats,
                                                 var_format=var_format)
            if legend_stats in [put.LegendStats.LAST, put.LegendStats.FIRST_LAST_NON_ZERO]:
                legend_title = f"Total: last={var_format.format(re_indexed_data.sum(axis=1).iloc[-1])}"

        put.set_legend(ax=ax,
                       labels=legend_labels,
                       colors=colors,
                       legend_loc=legend_loc,
                       reverse_columns=reverse_columns,
                       ncols=ncols,
                       bbox_to_anchor=bbox_to_anchor,
                       fontsize=fontsize,
                       legend_title=legend_title,
                       **kwargs)

    if title is not None:
        put.set_title(ax=ax, title=title, fontsize=fontsize, **kwargs)

    put.set_spines(ax=ax, **kwargs)

    return fig


def create_sample_portfolio_data() -> pd.DataFrame:
    """Create sample portfolio data for demonstration"""
    dates = pd.date_range('2010-01-01', periods=120, freq='ME')

    # Sample portfolio weights (can be positive or negative)
    np.random.seed(42)
    assets = ['AAPL', 'GOOGL', 'MSFT', 'TSLA', 'META']

    # Generate random weights that sum to approximately 100% (allowing for some leverage)
    weights_data = []
    for i in range(len(dates)):
        weights = np.random.normal(20, 15, len(assets))  # Can be negative (short positions)
        weights = weights / weights.sum() * 100  # Normalize to sum to 100%
        weights_data.append(weights)

    weights_df = pd.DataFrame(weights_data, columns=assets, index=dates)
    return weights_df


class LocalTests(Enum):
    WEIGHTS = 1


def run_local_test(local_test: LocalTests):
    """Run local tests for development and debugging purposes.

    These are integration tests that download real data and generate reports.
    Use for quick verification during development.
    """
    if local_test == LocalTests.WEIGHTS:

        weights = create_sample_portfolio_data()

        fig, axs = plt.subplots(2, 1, figsize=(8, 10), tight_layout=True)
        plot_stack(df=weights,
                   stacked=False,
                   use_bar_plot=True,
                   x_rotation=90,
                   yvar_format='{:,.0%}',
                   date_format='%b-%y',
                   fontsize=6,
                   ax=axs[0])

        plot_stack(df=weights,
                   stacked=False,
                   use_bar_plot=False,
                   x_rotation=90,
                   yvar_format='{:,.0%}',
                   date_format='%b-%y',
                   fontsize=6,
                   ax=axs[1])

    plt.show()


if __name__ == '__main__':

    run_local_test(local_test=LocalTests.WEIGHTS)



"""
quantile-quantile plot
"""
# packages
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from scipy import stats as stats
from statsmodels import api as sm
from typing import List, Union, Tuple, Optional
from enum import Enum

# qis
import qis.perfstats.returns as ret
import qis.plots.utils as put
import qis.perfstats.desc_table as dsc


def plot_qq(df: Union[pd.DataFrame, pd.Series],
            colors: List[str] = None,
            markers: List[str] = None,
            legend_loc: str = 'upper left',
            var_format: str = '{:.2f}',
            is_drop_na: bool = True,
            fontsize: int = 10,
            markersize: int = 2,
            title: str = None,
            xlabel: str = 'Theoretical quantiles',
            ylabel: str = 'Empirical quantiles',
            desc_table_type: dsc.DescTableType = dsc.DescTableType.SHORT,
            legend_stats: put.LegendStats = put.LegendStats.NONE,
            x_limits: Tuple[Optional[float], Optional[float]] = None,
            y_limits: Tuple[Optional[float], Optional[float]] = None,
            ax: plt.Subplot = None,
            **kwargs
            ) -> plt.Figure:

    if ax is None:
        fig, ax = plt.subplots()
    else:
        fig = None

    if isinstance(df, pd.Series):
        df = df.to_frame()
        line = 'q'
    else:
        line= None

    if colors is None:
         colors = put.get_n_colors(n=len(df.columns), **kwargs)

    if markers is None:
        markers = len(df.columns) * ['o']

    for idx, column in enumerate(df.columns):
        data0 = df[column]
        if is_drop_na:
            data0 = data0.dropna()
        sm.qqplot(data0, stats.norm, fit=True, line=line, ax=ax,  fmt=colors[idx],
                  markerfacecolor=colors[idx], markeredgecolor=colors[idx], marker=markers[idx],
                  markersize=markersize)
    if line is None:
        sm.qqline(ax, line='45', fmt='-', color='red')

    if desc_table_type != dsc.DescTableType.NONE:
        stats_table = dsc.compute_desc_table(df=df,
                                         desc_table_type=desc_table_type,
                                         var_format=var_format)
        put.set_legend_with_stats_table(stats_table=stats_table,
                                        ax=ax,
                                        colors=colors,
                                        legend_loc=legend_loc,
                                        fontsize=fontsize,
                                        **kwargs)
    else:
        legend_labels = put.get_legend_lines(data=df,
                                             legend_stats=legend_stats,
                                             var_format=var_format)
        put.set_legend(ax=ax,
                       labels=legend_labels,
                       colors=colors,
                       legend_loc=legend_loc,
                       fontsize=fontsize,
                       **kwargs)

    put.set_ax_xy_labels(ax=ax, xlabel=xlabel, ylabel=ylabel, fontsize=fontsize, **kwargs)
    put.set_ax_ticks_format(ax=ax, xvar_format=var_format, yvar_format=var_format, fontsize=fontsize, **kwargs)
    put.set_spines(ax=ax, **kwargs)

    if y_limits is not None:
        put.set_y_limits(ax=ax, y_limits=y_limits)
    if x_limits is not None:
        put.set_x_limits(ax=ax, x_limits=x_limits)

    if title is not None:
        put.set_title(ax=ax, title=title, fontsize=fontsize)

    return fig


def plot_xy_qq(x: pd.Series,
               y: pd.Series,
               colors: List[str] = None,
               markers: List[str] = None,
               labels: List[str] = None,
               legend_loc: str = 'upper left',
               is_drop_na: bool = True,
               ax: plt.Subplot = None,
               **kwargs
               ) -> plt.Figure:

    if ax is None:
        fig, ax = plt.subplots()
    else:
        fig = None

    if colors is None:
         colors = put.get_n_colors(n=1)

    if is_drop_na:
        x = x.dropna()
        y = y.dropna()
    x = x.to_numpy()
    y = y.to_numpy()

    qs = np.linspace(0, 1, min(len(x), len(y)))

    x_qs = np.quantile(x, qs)
    y_qs = np.quantile(y, qs)
    ax.scatter(x_qs, y_qs, c=colors[0])

    sm.qqline(ax, line='45', fmt='k--')

    put.set_legend(ax=ax,
                   labels=labels,
                   colors=colors,
                   legend_loc=legend_loc,
                   **kwargs)

    put.set_ax_xy_labels(ax=ax, **kwargs)
    put.set_ax_ticks_format(ax=ax, **kwargs)

    return fig


class LocalTests(Enum):
    RETURNS = 1
    XY_PLOT = 2


def run_local_test(local_test: LocalTests):
    """Run local tests for development and debugging purposes.

    These are integration tests that download real data and generate reports.
    Use for quick verification during development.
    """

    from qis.test_data import load_etf_data
    prices = load_etf_data().dropna()

    df = ret.to_returns(prices=prices, drop_first=True)

    if local_test == LocalTests.RETURNS:
        fig, ax = plt.subplots(1, 1, figsize=(8, 6))
        global_kwargs = dict(fontsize=8, linewidth=0.5, weight='normal', markersize=1)

        plot_qq(df=df,
                desc_table_type=dsc.DescTableType.SKEW_KURTOSIS,
                ax=ax,
                **global_kwargs)

    elif local_test == LocalTests.XY_PLOT:
        fig, ax = plt.subplots(1, 1, figsize=(8, 6))
        global_kwargs = dict(fontsize=8, linewidth=0.5, weight='normal', markersize=1)
        plot_xy_qq(x=df.iloc[:, 1],
                   y=df.iloc[:, 0],
                   ax=ax,
                   **global_kwargs)

    plt.show()


if __name__ == '__main__':

    run_local_test(local_test=LocalTests.RETURNS)




from qis.plots.utils import (
    TrendLine,
    LastLabel,
    LegendStats,
    add_scatter_points,
    align_x_limits_ax12,
    align_x_limits_axs,
    align_xy_limits,
    align_y_limits_ax12,
    align_y_limits_axs,
    autolabel,
    compute_heatmap_colors,
    create_dummy_line,
    get_cmap_colors,
    get_data_group_colors,
    get_legend_lines,
    get_n_cmap_colors,
    get_n_colors,
    get_n_fixed_colors,
    get_n_hatch,
    get_n_markers,
    get_n_mlt_colors,
    get_n_sns_colors,
    map_dates_index_to_str,
    rand_cmap,
    remove_spines,
    set_ax_tick_labels,
    set_ax_tick_params,
    set_ax_ticks_format,
    set_ax_xy_labels,
    set_date_on_axis,
    set_legend,
    set_legend_colors,
    set_legend_with_stats_table,
    set_linestyles,
    set_spines,
    set_suptitle,
    set_title,
    set_x_limits,
    set_y_limits,
    subplot_border,
    validate_returns_plot,
    calc_table_height,
    calc_table_width,
    calc_df_table_size,
    get_df_table_size,
    reset_xticks,
    set_labels_frequency,
    scale_ax_bar_width,
    get_table_lines_for_group_data
)

from qis.plots.bars import plot_bars, plot_vbars

from qis.plots.boxplot import (
    plot_box,
    df_boxplot_by_classification_var,
    df_boxplot_by_hue_var,
    df_boxplot_by_index,
    df_boxplot_by_columns,
    df_dict_boxplot_by_columns,
    df_dict_boxplot_by_classification_var
)

from qis.plots.contour import plot_contour

from qis.plots.errorbar import plot_errorbar

from qis.plots.heatmap import plot_heatmap

from qis.plots.histogram import plot_histogram, PdfType

from qis.plots.histplot2d import plot_histplot2d

from qis.plots.lineplot import plot_line, plot_lines_list

from qis.plots.pie import plot_pie

from qis.plots.qqplot import plot_qq, plot_xy_qq

from qis.plots.scatter import (plot_scatter,
                               plot_classification_scatter,
                               plot_multivariate_scatter_with_prediction)

from qis.plots.stackplot import plot_stack

from qis.plots.table import (
    plot_df_table,
    plot_df_table_with_ci,
    set_align_for_column,
    set_cells_facecolor,
    set_column_edge_color,
    set_data_colors,
    set_diag_cells_facecolor,
    set_row_edge_color
)

from qis.plots.time_series import (
    plot_time_series,
    plot_time_series_2ax
)


from qis.plots.derived.prices import (
    add_bnb_regime_shadows,
    get_performance_labels_for_stats,
    get_performance_labels_for_stats,
    PerfStatsLabels,
    plot_prices,
    plot_prices_2ax,
    plot_prices_with_dd,
    plot_prices_with_fundamentals,
    plot_rolling_perf_stat
)

from qis.plots.derived.data_timeseries import plot_data_timeseries


from qis.plots.derived.perf_table import (
    plot_desc_freq_table,
    plot_ra_perf_annual_matrix,
    plot_ra_perf_bars,
    plot_ra_perf_by_dates,
    plot_ra_perf_scatter,
    get_ra_perf_columns,
    plot_ra_perf_table,
    plot_ra_perf_table_benchmark,
    get_ra_perf_benchmark_columns,
    plot_top_bottom_performers,
    plot_best_worst_returns
)

from qis.plots.derived.regime_class_table import get_quantile_class_table, plot_quantile_class_table

from qis.plots.derived.regime_pdf import plot_regime_pdf

from qis.plots.derived.regime_scatter import plot_scatter_regression

from qis.plots.derived.returns_heatmap import (
    compute_periodic_returns_by_row_table,
    compute_periodic_returns_table,
    compute_periodic_returns,
    plot_periodic_returns_table,
    plot_returns_heatmap,
    plot_returns_table,
    plot_sorted_periodic_returns
)

from qis.plots.derived.returns_scatter import plot_returns_scatter


from qis.plots.derived.drawdowns import (
    DdLegendType,
    plot_rolling_drawdowns,
    plot_rolling_time_under_water,
    plot_top_drawdowns_paths
)

from qis.plots.derived.regime_data import (
    plot_regime_data,
    plot_regime_boxplot,
    add_bnb_regime_shadows
)

from qis.plots.derived.desc_table import plot_desc_table

from qis.plots.reports.utils import ReportType

from qis.plots.reports.econ_data_single import econ_data_report

from qis.plots.reports.price_history import (plot_price_history,
                                             generate_price_history_report)



import matplotlib.pyplot as plt
from enum import Enum
import yfinance as yf
import qis as qis


class LocalTests(Enum):
    CORR1 = 1


def run_local_test(local_test: LocalTests):
    """Run local tests for development and debugging purposes.

    These are integration tests that download real data and generate reports.
    Use for quick verification during development.
    """

    tickers = ['SPY', 'QQQ', 'TLT', 'GLD']
    prices = yf.download(tickers, start="2003-12-31", end=None, ignore_tz=True, auto_adjust=True)['Close'][tickers].dropna()

    if local_test == LocalTests.CORR1:
        fig, ax = plt.subplots(1, 1, figsize=(7, 7), tight_layout=True)
        qis.plot_returns_ewm_corr_table(prices=prices,
                                        ewm_lambda=0.97,
                                        ax=ax)
    plt.show()


if __name__ == '__main__':

    run_local_test(local_test=LocalTests.CORR1)



"""
factsheet for multi strategy report for cross sectional comparison of strategies
and generating sensitivities to parameters
see example in qis.examples.factheets.multi_strategy.py
"""
# packages
import pandas as pd
import matplotlib.pyplot as plt
from typing import Tuple, List
# qis
import qis as qis
from qis import TimePeriod, PerfParams, PerfStat, BenchmarkReturnsQuantilesRegime
from qis.portfolio.multi_portfolio_data import MultiPortfolioData
from qis.portfolio.reports.config import PERF_PARAMS, regime_classifier
from qis.portfolio.reports.strategy_factsheet import generate_strategy_factsheet


def generate_multi_portfolio_factsheet(multi_portfolio_data: MultiPortfolioData,
                                       time_period: TimePeriod = None,
                                       perf_params: PerfParams = PERF_PARAMS,
                                       regime_classifier: BenchmarkReturnsQuantilesRegime = BenchmarkReturnsQuantilesRegime(),
                                       regime_benchmark: str = None,
                                       backtest_name: str = None,
                                       heatmap_freq: str = 'YE',
                                       add_benchmarks_to_navs: bool = False,
                                       figsize: Tuple[float, float] = (8.3, 11.7),  # A4 for portrait
                                       group_data: pd.Series = None,
                                       add_group_exposures_and_pnl: bool = False,
                                       add_strategy_factsheets: bool = False,
                                       fontsize: int = 5,
                                       **kwargs
                                       ) -> List[plt.Figure]:
    """
    for portfolio data with structurally different strategies
    for portfolios with large universe use is_grouped = True to report turnover and exposures by groups
    """
    if group_data is not None:
        is_grouped = True
    else:
        is_grouped = False

    if regime_benchmark is None:
        regime_benchmark = multi_portfolio_data.benchmark_prices.columns[0]

    plot_kwargs = dict(fontsize=fontsize,
                       linewidth=0.5,
                       digits_to_show=1, sharpe_digits=2,
                       weight='normal',
                       markersize=1,
                       framealpha=0.75)
    kwargs = qis.update_kwargs(kwargs, plot_kwargs)

    fig = plt.figure(figsize=figsize, constrained_layout=True)
    gs = fig.add_gridspec(nrows=7, ncols=4, wspace=0.0, hspace=0.0)

    if backtest_name is not None:
        fig.suptitle(backtest_name, fontweight="bold", fontsize=8, color='blue')

    multi_portfolio_data.plot_nav(ax=fig.add_subplot(gs[0, :2]),
                                  time_period=time_period,
                                  regime_benchmark=regime_benchmark,
                                  perf_params=perf_params,
                                  regime_classifier=regime_classifier,
                                  add_benchmarks_to_navs=add_benchmarks_to_navs,
                                  title=f"Cumulative performance with background colors using bear/normal/bull regimes of {regime_benchmark} {regime_classifier.freq}-returns",
                                  **kwargs)

    multi_portfolio_data.plot_drawdowns(ax=fig.add_subplot(gs[1, :2]),
                                        time_period=time_period,
                                        regime_benchmark=regime_benchmark,
                                        regime_classifier=regime_classifier,
                                        dd_legend_type=qis.DdLegendType.SIMPLE,
                                        add_benchmarks_to_navs=add_benchmarks_to_navs,
                                        title='Running Drawdowns',
                                        **kwargs)

    multi_portfolio_data.plot_rolling_time_under_water(ax=fig.add_subplot(gs[2, :2]),
                                                       time_period=time_period,
                                                       regime_benchmark=regime_benchmark,
                                                       regime_classifier=regime_classifier,
                                                       add_benchmarks_to_navs=add_benchmarks_to_navs,
                                                       title='Rolling time under water',
                                                       **kwargs)

    multi_portfolio_data.plot_rolling_perf(ax=fig.add_subplot(gs[3, :2]),
                                           time_period=time_period,
                                           regime_benchmark=regime_benchmark,
                                           regime_classifier=regime_classifier,
                                           add_benchmarks_to_navs=add_benchmarks_to_navs,
                                           **kwargs)

    multi_portfolio_data.plot_exposures(ax=fig.add_subplot(gs[4, :2]),
                                        portfolio_idx=0,
                                        time_period=time_period,
                                        regime_benchmark=regime_benchmark,
                                        regime_classifier=regime_classifier,
                                        **kwargs)

    multi_portfolio_data.plot_turnover(ax=fig.add_subplot(gs[5, :2]),
                                       time_period=time_period,
                                       regime_benchmark=regime_benchmark,
                                       regime_classifier=regime_classifier,
                                       **kwargs)

    multi_portfolio_data.plot_costs(ax=fig.add_subplot(gs[6, :2]),
                                    time_period=time_period,
                                    regime_benchmark=regime_benchmark,
                                    regime_classifier=regime_classifier,
                                    **kwargs)

    multi_portfolio_data.plot_performance_bars(ax=fig.add_subplot(gs[0, 2]),
                                               perf_params=perf_params,
                                               perf_column=PerfStat.SHARPE_RF0,
                                               time_period=time_period,
                                               add_benchmarks_to_navs=add_benchmarks_to_navs,
                                               **qis.update_kwargs(kwargs, dict(fontsize=fontsize)))

    multi_portfolio_data.plot_performance_bars(ax=fig.add_subplot(gs[0, 3]),
                                               perf_params=perf_params,
                                               perf_column=PerfStat.MAX_DD,
                                               time_period=time_period,
                                               add_benchmarks_to_navs=add_benchmarks_to_navs,
                                               **qis.update_kwargs(kwargs, dict(fontsize=fontsize)))

    multi_portfolio_data.plot_ra_perf_table(ax=fig.add_subplot(gs[1, 2:]),
                                            perf_params=perf_params,
                                            time_period=time_period,
                                            **qis.update_kwargs(kwargs, dict(fontsize=fontsize)))

    multi_portfolio_data.plot_periodic_returns(ax=fig.add_subplot(gs[2, 2:]),
                                               heatmap_freq=heatmap_freq,
                                               title=f"{heatmap_freq} returns",
                                               time_period=time_period,
                                               add_benchmarks_to_navs=add_benchmarks_to_navs,
                                               **qis.update_kwargs(kwargs, dict(fontsize=fontsize)))

    """
    multi_portfolio_data.plot_ra_perf_table(ax=fig.add_subplot(gs[3, 2:]),
                                            perf_params=perf_params,
                                            time_period=qis.get_time_period_shifted_by_years(time_period=time_period),
                                            **qis.update_kwargs(kwargs, dict(fontsize=fontsize, freq_reg='W-WED')))
    """
    multi_portfolio_data.plot_corr_table(ax=fig.add_subplot(gs[3, 2:]),
                                         time_period=time_period,
                                         freq=perf_params.freq,
                                         add_benchmarks_to_navs=add_benchmarks_to_navs,
                                         **qis.update_kwargs(kwargs, dict(fontsize=fontsize)))

    if len(multi_portfolio_data.benchmark_prices.columns) > 1:
        multi_portfolio_data.plot_regime_data(ax=fig.add_subplot(gs[4, 2]),
                                              is_grouped=is_grouped,
                                              time_period=time_period,
                                              perf_params=perf_params,
                                              regime_classifier=regime_classifier,
                                              benchmark=multi_portfolio_data.benchmark_prices.columns[0],
                                              **kwargs)
        multi_portfolio_data.plot_regime_data(ax=fig.add_subplot(gs[4, 3]),
                                              is_grouped=is_grouped,
                                              time_period=time_period,
                                              perf_params=perf_params,
                                              regime_classifier=regime_classifier,
                                              benchmark=multi_portfolio_data.benchmark_prices.columns[1],
                                              **kwargs)
    else:
        multi_portfolio_data.plot_regime_data(ax=fig.add_subplot(gs[4, 2:]),
                                              is_grouped=is_grouped,
                                              time_period=time_period,
                                              perf_params=perf_params,
                                              regime_classifier=regime_classifier,
                                              benchmark=multi_portfolio_data.benchmark_prices.columns[0],
                                              **kwargs)
    if len(multi_portfolio_data.benchmark_prices.columns) > 1:
        # take first two benchmarks
        benchmark_prices = multi_portfolio_data.benchmark_prices.iloc[:, :2]
        multi_portfolio_data.plot_factor_betas(axs=[fig.add_subplot(gs[5, 2:]), fig.add_subplot(gs[6, 2:])],
                                               benchmark_prices=benchmark_prices,
                                               time_period=time_period,
                                               regime_benchmark=regime_benchmark,
                                               regime_classifier=regime_classifier,
                                               **kwargs)
    else:
        multi_portfolio_data.plot_returns_scatter(ax=fig.add_subplot(gs[5, 2:]),
                                                  time_period=time_period,
                                                  benchmark=multi_portfolio_data.benchmark_prices.columns[0],
                                                  **qis.update_kwargs(kwargs, dict(freq=perf_params.freq_reg)))

        multi_portfolio_data.plot_factor_betas(axs=[fig.add_subplot(gs[6, 2:])],
                                               time_period=time_period,
                                               benchmark_prices=multi_portfolio_data.benchmark_prices,
                                               regime_benchmark=regime_benchmark,
                                               regime_classifier=regime_classifier,
                                               **kwargs)

    figs = [fig]
    if add_group_exposures_and_pnl:
        figs1 = multi_portfolio_data.plot_group_exposures_and_pnl(time_period=time_period,
                                                                  regime_benchmark=regime_benchmark,
                                                                  regime_classifier=regime_classifier,
                                                                  **kwargs)
        figs.append(figs1)

    if add_strategy_factsheets:
        for portfolio_data in multi_portfolio_data.portfolio_datas:
            figs.append(generate_strategy_factsheet(portfolio_data=portfolio_data,
                                                    benchmark_prices=multi_portfolio_data.benchmark_prices,
                                                    perf_params=perf_params,
                                                    regime_classifier=regime_classifier,
                                                    add_grouped_exposures=is_grouped,
                                                    time_period=time_period,
                                                    **kwargs
                                                    ))
    figs = qis.to_flat_list(figs)
    return figs



"""
Example of using multivariate ewma betas
"""
import seaborn as sns
import matplotlib.pyplot as plt
import yfinance as yf
import qis as qis
from qis import EwmLinearModel

tickers = ['SPY', 'TLT', 'GLD']
prices = yf.download(tickers=tickers, start="2003-12-31", end=None, ignore_tz=True, auto_adjust=True)['Close'][tickers]
prices = prices.asfreq('B', method='ffill').dropna()  # make B frequency and align
returns = qis.to_returns(prices, is_log_returns=True, drop_first=True, freq='ME')
x = returns[['SPY', 'TLT']]
y = returns[['GLD']]


ewm_linear_model = EwmLinearModel(x=x, y=y)
ewm_linear_model.fit(span=12, is_x_correlated=True, mean_adj_type=qis.MeanAdjType.EWMA)
betas = ewm_linear_model.get_asset_factor_betas(asset=y.columns[0])
print(betas)

with sns.axes_style("darkgrid"):
    fig, ax = plt.subplots(1, 1, figsize=(12, 6))
    qis.plot_time_series(df=betas, ax=ax)

plt.show()



# packages
import pandas as pd
import matplotlib.pyplot as plt
import qis as qis
from enum import Enum
from bbg_fetch import fetch_field_timeseries_per_tickers


def run_report():
    # tickers = {'TY1 Comdty': '10y', 'UXY1 Comdty': '10y Ultra'}
    tickers = {
        'SPTR Index': 'SPTR Index',
        'CIEQVEHG Index': 'Citi SPX 0D Vol Carry',
        'CIEQVRUG Index': 'Citi SX5E 1W Vol Carry',
        'CICXCOSE  Index': 'Citi Brent Vol Carry',
        'GSISXC07 Index': 'GS Multi Asset Carry',
        'GSISXC11 Index': 'GS Macro Carry',
        'XUBSPGRA Index': 'UBS Gold Strangles',
        'XUBSU1D1 Index': 'UBS Short Vol Daily',
        'BCKTARU2 Index': 'BNP Call on Short-vol Carry',
        'BNPXAUUS Index': 'BNP Intraday SPX Vol Carry',
        'BNPXAUTS Index': 'BNP Intraday NDX Vol Carry',
        'BNPXOV3U Index': 'BNP 3M Long DHhedged Puts'
        }

    benchmark = 'HYG US Equity'
    tickers = {
        benchmark: benchmark,
        'NMVVR1EL Index': 'IRVING1 EUR',
        'NMVVR1UL Index': 'IRVING1 USD',
        'NMVVR1L Index': 'IRVING1',
        'BNPXLVRE Index': 'BNP Long Rates Vol EUR',
        'BNPXLVRU Index': 'BNP Long Rates Vol USD',
        'BXIIULSV Index': 'Barclays Long Rates Vol',
        'BXIIUGNT Index': 'Barclays Gamma Neutral Vol',
        'BXIIUENT Index': 'Barclays Triangle Vol'
    }
    
    benchmark = 'SPTR Index'
    tickers = {
        benchmark: benchmark,
        'BNPIV1EE Index': 'BNP Europe 1Y Volatility',
        'BNPIV1UE Index': 'BNP US 1Y Volatility',
        'BNPXVO3A Index': 'BNP VOLA 3 Index',
        'AIJPVT1U Index': 'JPM Volatility Trend Following',
        'JPOSLVUS Index': 'JPM US Long Variance',
        'JPOSPRU2 Index': 'JPM US Put Ratio',
        'JPOSTUDN Index': 'JPM US Equity Tail Hedge',
        'JPRC85BE Index': 'JPM Dynamic 85% Rolling Collar EU',
        'JPRC85BU Index': 'JPM Dynamic 85% Rolling Collar US',
        'JPUSVXCR Index': 'JPM US Volatility Call Ratio'
    }

    benchmark = 'XNDX Index'
    tickers = {
        benchmark: benchmark,
        'BNPXTHUE Index': 'Thalia',
        'BNPXTHUN Index': 'Thalia Neutral',
        'BNPXTDUE Index': 'Thalia Dynamic',
        'BNPXTDUN Index': 'Thalia Neutral Dynamic',
        'BNPXLVRU Index': 'BNP Long Rates Vol USD'
    }

    benchmark = 'SPTR Index'
    tickers = {
        benchmark: benchmark,
        'AIJPMT1U Index': 'JPM Macro Trend',
        'AIJPLT3U Index': 'JPM Cross Trend',
        'AIJPXSK1 Index': 'JPM XA Skeweness',
        'NEIXCTAT Index': 'SG Trend'
    }

    benchmark = 'SPTR Index'
    tickers = {
        benchmark: benchmark,
        'JPQGM4W1 Index': 'JPM Factor1',
        'JPQTR4W1 Index': 'JPM Factor2'
    }

    benchmark = 'SPTR Index'
    tickers = {
        benchmark: benchmark,
        'DBBNE05Y Index': 'DBBNE05Y',
        'DBBNE10Y Index': 'DBBNE10Y',
        'DBBNE15Y Index': 'DBBNE15Y',
        'DBBNU05Y Index': 'DBBNU05Y',
        'DBCUU10Y Index': 'DBCUU10Y',
        'DBBNU15Y Index': 'DBBNU15Y'
    }

    benchmark = 'SPTR Index'
    tickers = {
        benchmark: benchmark,
        'CICMCI5B Index': 'CDX IG Citi',
        'UISYMI5S Index': 'CDX IG UBS shortable',
        'DBCDIG5F Index': 'CDX IG DB long fixed',
        'DBCDIG5L Index': 'CDX IG DB long variable',
        'DBCDIG5S Index': 'CDX IG DB short',
        'CICMCH5B Index': 'CDX HY Citi',
        'UISYMH5S Index': 'CDX HY UBS shortable',
        'DBCDHYLG Index': 'CDX HY DB long fixed',
        'DBCDHY5A Index': 'CDX HY DB long variable',
        # 'DBCDHY5S Index': 'CDX HY DB short'
    }

    benchmark = 'AOR US Equity'
    tickers = {
        benchmark: benchmark,
        'HFRXGL Index': 'HFRXGL Index',
        'HFRIFWI Index': 'HFRIFWI Index',
        'GMSGRMU ID Equity': 'Graham',
        'WINTFUI ID Equity': 'Winton',
        'BHMG LN Equity': 'Brevan',
        'OMEIUSA ID Equity': 'Jupiter'
    }

    benchmark = 'SPTGGUT Index'
    tickers = {
        benchmark: benchmark,
        'NEIXCTAT Index': 'SG Trend',
        'NEIXCTA Index': 'SG CTA',
        'HFRIMTF Index': 'HFRI TF',
        'BHCTA Index': 'BBG CTA',
        'HFRXM Index': 'HFRX Macro/CTA'
    }


    prices = fetch_field_timeseries_per_tickers(tickers=tickers, freq='B', field='PX_LAST').ffill()
    print(prices)
    # qis.save_df_to_csv(df=prices, file_name='qis_vol_indices', local_path=qis.get_output_path())

    time_period = qis.TimePeriod('31Dec2014', '31Oct2025')
    # kwargs = qis.fetch_default_report_kwargs(time_period=time_period, add_rates_data=False)
    # kwargs = qis.fetch_factsheet_config_kwargs(factsheet_config=qis.FACTSHEET_CONFIG_DAILY_DATA_SHORT_PERIOD, add_rates_data=False)
    kwargs = qis.fetch_factsheet_config_kwargs(factsheet_config=qis.FACTSHEET_CONFIG_MONTHLY_DATA_LONG_PERIOD, add_rates_data=False)

    fig = qis.generate_multi_asset_factsheet(prices=prices,
                                             benchmark=benchmark,
                                             time_period=time_period,
                                             **kwargs)
    qis.save_figs_to_pdf(figs=[fig],
                         file_name=f"bbg_multiasset_report", orientation='landscape',
                         # local_path='C://Users//uarts//outputs//',
                         local_path=qis.get_output_path()
                         )


def run_price():
    tickers = {'CL1 Comdty': 'WTI'}
    prices = fetch_field_timeseries_per_tickers(tickers=tickers, freq='B', field='PX_LAST').ffill()
    print(prices)

    time_period = qis.TimePeriod('31Dec1989', '08Nov2024')
    prices = time_period.locate(prices)

    qis.plot_prices_with_dd(prices,
                            start_to_one=False)


class LocalTests(Enum):
    REPORT = 1
    PRICE = 2


def run_local_test(local_test: LocalTests):
    """Run local tests for development and debugging purposes.

    These are integration tests that download real data and generate reports.
    Use for quick verification during development.
    """
    pd.set_option('display.max_rows', 500)
    pd.set_option('display.max_columns', 500)
    pd.set_option('display.width', 1000)


    if local_test == LocalTests.REPORT:
        run_report()

    elif local_test == LocalTests.PRICE:
        run_price()

    plt.show()


if __name__ == '__main__':

    run_local_test(local_test=LocalTests.REPORT)
